import matplotlib.pyplot as plt
import matplotlib
import numpy as np

# written on 10/16/2017 added a function for making prettier plots for presentations
# 4/20/2019 added in a function to add in more interactions for matplotlib plots


def presentation_plot():
    # function for setting matplotlib to makes plots that are actually visible for presentation and papers
    
    #set font size for plotting
    matplotlib.rcParams.update({'font.size': 16})
    #set axes linewidth for plotting
    matplotlib.rcParams['axes.linewidth'] = 2
    # set tick width
    matplotlib.rcParams['xtick.major.size'] = 10
    matplotlib.rcParams['xtick.direction'] = "in"
    matplotlib.rcParams['ytick.direction'] = "in"
    matplotlib.rcParams['xtick.major.width'] = 2
    matplotlib.rcParams['xtick.minor.size'] = 5
    matplotlib.rcParams['xtick.minor.width'] = 1
    matplotlib.rcParams['ytick.major.size'] = 10
    matplotlib.rcParams['ytick.major.width'] = 2
    matplotlib.rcParams['ytick.minor.size'] = 5
    matplotlib.rcParams['ytick.minor.width'] = 1



# function to add some functionality to matplotlib interactive plotting
# adds in the ability to pan around with the arrow keys
# adds in the ability to zoom in and out using the keyboard keys z and x
# adds in the ability to grab cursur info from screen
# adds in the ability to find mins and maximums 
def activate_interactive_plotting(fig,ax,zoom_factor = 0.1,lim_shift_factor = 0.1,show_pts = True):   
    class InteractivePlot(object): #must pass fig to this command
    
        def __init__(self,fig,ax,zoom_factor,lim_shift_factor,show_pts):
            #matplotlib.use("TkAgg")
            #print("hello world")
            self.fig = fig
            self.ax = ax
            self.show_pts = show_pts
            self.x_data = np.asarray(())
            self.y_data = np.asarray(())
            self.pts, = self.ax.plot(self.x_data,self.y_data,"+")
            self.text_dict = {}
            plt.rcParams['keymap.forward'] = ['v']
            plt.rcParams['keymap.back'] = ['c','backspace']# remove arrows from back and forward on plot
            self.fig.canvas.mpl_connect('key_press_event', self.on_key_press)
            self.fig.canvas.mpl_connect('key_press_event', self.on_key_press)
            self.fig.canvas.mpl_connect('key_release_event', self.on_key_release)
            self.fig.canvas.mpl_connect('button_press_event', self.onClick)
            self.shift_is_held = False
            self.control_is_held = False
            self.alt_is_held = False
            self.lim_shift_factor = lim_shift_factor
            self.zoom_factor = zoom_factor #no greater than 0.5
            print("")
            print("Interactive Plotting Tools Activated")
            print("Use arrow keys to pan.")
            print("Use z and x keys to zoom and Xplode.")
            print("Hold shift and right click to get cursur location data printed to screen")
            print("   and stored to x_data, y_data arrays.")
            print("Hold control and right click to remove a grabed data point from the plot")
            print("   and remove it from x_data, y_data arrays.")
            print("Hold alt and right click to get the nearest data point for plotted data")
            print("   and store it to x_data, y_data arrays")
            print("Press m to find the minimum point currently displayed")
            print("Press n to find the maximum point currently displayed")
            print("On mac shift and control key requires matplotlib backend TkAgg.")
            print("")

            #plt.show(block = True)

        def on_key_press(self, event):
            #print( event.key)
            if event.key == 'shift':
                self.shift_is_held = True
            if event.key == 'control':
                self.control_is_held = True
            if event.key == 'alt':
                self.alt_is_held = True
               
            if event.key == 'right': #pan right
                xlim_left, xlim_right = self.ax.get_xlim() 
                xlim_size = xlim_right-xlim_left
                self.ax.set_xlim(xlim_left+self.lim_shift_factor*xlim_size,xlim_right+self.lim_shift_factor*xlim_size)
                self.fig.canvas.draw()

            if event.key == 'left': #pan left
                xlim_left, xlim_right = self.ax.get_xlim() 
                xlim_size = xlim_right-xlim_left
                self.ax.set_xlim(xlim_left-self.lim_shift_factor*xlim_size,xlim_right-self.lim_shift_factor*xlim_size)
                self.fig.canvas.draw()

            if event.key == 'up': #pan up
                ylim_left, ylim_right = self.ax.get_ylim() 
                ylim_size = ylim_right-ylim_left
                self.ax.set_ylim(ylim_left+self.lim_shift_factor*ylim_size,ylim_right+self.lim_shift_factor*ylim_size)
                self.fig.canvas.draw()

            if event.key == 'down': #pan down
                ylim_left, ylim_right = self.ax.get_ylim() 
                ylim_size = ylim_right-ylim_left
                self.ax.set_ylim(ylim_left-self.lim_shift_factor*ylim_size,ylim_right-self.lim_shift_factor*ylim_size)
                self.fig.canvas.draw()

            if event.key == 'z': #zoom in
                xlim_left, xlim_right = self.ax.get_xlim() 
                ylim_left, ylim_right = self.ax.get_ylim() 
                xlim_size = xlim_right-xlim_left
                ylim_size = ylim_right-ylim_left
                self.ax.set_xlim(xlim_left+self.zoom_factor*xlim_size,xlim_right-self.zoom_factor*xlim_size)
                self.ax.set_ylim(ylim_left+self.zoom_factor*ylim_size,ylim_right-self.zoom_factor*ylim_size)
                self.fig.canvas.draw()
            elif event.key == 'Z':
                print("Is caps lock on?")

            if event.key == 'x': #zoom out
                xlim_left, xlim_right = self.ax.get_xlim() 
                ylim_left, ylim_right = self.ax.get_ylim() 
                xlim_size = xlim_right-xlim_left
                ylim_size = ylim_right-ylim_left
                self.ax.set_xlim(xlim_left-self.zoom_factor*xlim_size,xlim_right+self.zoom_factor*xlim_size)
                self.ax.set_ylim(ylim_left-self.zoom_factor*ylim_size,ylim_right+self.zoom_factor*ylim_size)
            elif event.key == 'X':
                print("Is caps lock on?")

            if event.key == 'm':
                self.find_min()
            elif event.key == 'M':
                print("Is caps lock on?")

            if event.key == 'n':
                self.find_max()
            elif event.key == 'N':
                print("Is caps lock on?")

        def on_key_release(self, event):
            if event.key == 'shift':
                self.shift_is_held = False
            if event.key == 'control':
                self.control_is_held = False
            if event.key == 'alt':
                self.alt_is_held = False

        def onClick(self, event):
            if event.button == 3:
                if self.alt_is_held: # add point on line
                    xlim_left, xlim_right = self.ax.get_xlim() # only look in current axis
                    index_low = np.argmin(np.abs(self.ax.lines[0].get_xdata()-xlim_left))
                    index_high  = np.argmin(np.abs(self.ax.lines[0].get_xdata()-xlim_right))+1
                    index = index_low + np.argmin(np.abs(self.ax.lines[0].get_xdata()[index_low:index_high]-event.xdata))
                    self.x_data = np.append(self.x_data,self.ax.lines[0].get_xdata()[index])
                    self.y_data = np.append(self.y_data,self.ax.lines[0].get_ydata()[index])
                    if self.show_pts == True:
                        self.refresh_plot()
                if self.shift_is_held: # add point
                    print("x, y = ", event.xdata,event.ydata)
                    self.x_data = np.append(self.x_data,event.xdata)
                    self.y_data = np.append(self.y_data,event.ydata)
                    if self.show_pts == True:
                        self.refresh_plot()
                if self.control_is_held: # delete point
                    if len(self.x_data)>0:
                        print("removing point")
                        delete_index = np.argmin(np.abs(self.x_data-event.xdata))
                        self.x_data = np.delete(self.x_data,delete_index)
                        self.y_data = np.delete(self.y_data,delete_index)
                        self.refresh_plot()



        def refresh_plot(self):
            self.pts.set_data(self.x_data,self.y_data)
            if len(self.text_dict)>0:
                for i in range(0,len(self.text_dict)):
                    self.text_dict[i].set_text("")# clear all of the texts
            self.text_dict = {} # empty the dictionary
            for i in range(0,len(self.x_data)): #rebuild the dictionary
                self.text_dict[i] = self.ax.text(self.x_data[i], self.y_data[i], str(i))
            self.fig.canvas.draw()

        def find_min(self):
            print("Finding minimum")
            xlim_left, xlim_right = self.ax.get_xlim() # only look in current axis
            index_low = np.argmin(np.abs(self.ax.lines[0].get_xdata()-xlim_left))
            index_high  = np.argmin(np.abs(self.ax.lines[0].get_xdata()-xlim_right))+1
            index = index_low + np.argmin(self.ax.lines[0].get_ydata()[index_low:index_high])
            self.x_data = np.append(self.x_data,self.ax.lines[0].get_xdata()[index])
            self.y_data = np.append(self.y_data,self.ax.lines[0].get_ydata()[index])
            self.refresh_plot()

        def find_max(self):
            print("Finding maximum")
            xlim_left, xlim_right = self.ax.get_xlim() # only look in current axis
            index_low = np.argmin(np.abs(self.ax.lines[0].get_xdata()-xlim_left))
            index_high  = np.argmin(np.abs(self.ax.lines[0].get_xdata()-xlim_right))+1
            index = index_low+np.argmax(self.ax.lines[0].get_ydata()[index_low:index_high])
            self.x_data = np.append(self.x_data,self.ax.lines[0].get_xdata()[index])
            self.y_data = np.append(self.y_data,self.ax.lines[0].get_ydata()[index])
            self.refresh_plot()

    
            

    ip = InteractivePlot(fig,ax,zoom_factor = zoom_factor,lim_shift_factor = lim_shift_factor,show_pts = show_pts)
    return ip





"""
Automated data storage and inspections tools for the various fitting functions
"""

import os
from operator import attrgetter
from typing import NamedTuple, Optional, Callable, Sequence

import toml
import numpy as np
import scipy.stats as stats
import matplotlib.pyplot as plt

from submm.sample_data.abs_paths import abs_path_output_dir_default
from submm.KIDs.res.utils import derived_text, filename_text, write_text, calc_qc_qi, line_format
from submm.KIDs.res.sweep_tools import InteractivePlot


derived_params = {'qi', 'qc', 'red_chi_sq', 'chi_sq', 'p_value'}
field_to_first_format_int = {'fr': 5, 'f_0': 5, 'f_min': 5, 'qr': 7, 'q': 7, 'q_e_real': 7, 'q_e_imag': 7, 'amp': 1, 'phi': 2,
                             'a': 1, 'a_mag': 1, 'a_slope': 2, 'b0': 2, 'b1': 2, 'i0': 1, 'q0': 1, 'tau': 6, 'delay': 6,
                             'f0': 5, 'qi': 7, 'qc': 7, 'flin': 5, 'red_chi_sq': 2, 'chi_sq': 1, 'p_value': 2}
field_to_decimal_format_int = {'fr': 4, 'f_0': 4,  'f_min': 4, 'qr': 0, 'q': 0, 'q_e_real': 0, 'q_e_imag': 0, 'amp': 2, 'phi': 2,
                               'a': 2, 'a_mag': 2, 'a_slope': 2, 'b0': 2, 'b1': 2, 'i0': 2, 'q0': 2, 'tau': 2, 'delay': 2,
                               'f0': 4, 'qi': 0, 'qc': 0, 'flin': 4, 'red_chi_sq': 3, 'chi_sq': 1,
                               'p_value': 3}
field_to_format_letter = {'fr': 'f', 'f_0': 'f',  'f_min': 'f', 'qr': 'f', 'q': 'f', 'q_e_real': 'f', 'q_e_imag': 'f',  'amp': 'f', 'phi': 'f',
                          'a': 'f', 'a_mag': 'f', 'a_slope': 'E', 'b0': 'E', 'b1': 'E', 'i0': 'E', 'q0': 'E', 'tau': 'f', 'delay': 'f',
                          'f0': 'f', 'qi': 'f', 'qc': 'f', 'flin': 'f', 'red_chi_sq': 'f',
                          'chi_sq': 'E',  'p_value': 'f'}

field_to_field_label = {'fr': 'fr (MHz)', 'f_0': 'f_O (MHz)',  'f_min': 'fmin (MHz)', 'qr': 'Qr', 'q': 'Q', 'q_e_real': 'Qe (real)', 'q_e_imag': 'Qe (imag)', 'amp': 'amp', 'phi': 'phi',
                        'a': 'a', 'a_mag': 'a_mag', 'a_slope': 'a_slope', 'b0': 'b0', 'b1': 'b1', 'i0': 'i0', 'q0': 'q0', 'tau': 'tau (ns)', 'delay': 'delay', 'f0': 'f0 (MHz)', 'qi': 'Qi', 'qc': 'Qc',
                        'flin': 'flin (MHz)', 'red_chi_sq': 'reduced chi-sqr', 'chi_sq': 'chi-sq', 'p_value': 'p-value'}
field_to_multiplier = {'fr': 1.0e-6, 'f_0': 1.0e-6, 'f_min': 1.0e-6, 'tau': 1.0e9, 'f0': 1.0e-6, 'flin': 1.0e-6}

field_to_format_strs = {}
field_to_text_len = {}
for a_field in field_to_first_format_int.keys():
    first_format_int = field_to_first_format_int[a_field]
    decimal_format_int = field_to_decimal_format_int[a_field]
    format_letter = field_to_format_letter[a_field]
    field_to_format_strs[a_field] = f'{first_format_int}.{decimal_format_int}{format_letter}'
    # the + 1 is to account for negative numbers with the extra '-'
    text_len = first_format_int + decimal_format_int + 1
    if decimal_format_int > 0:
        # add a space for the decimal point
        text_len += 1
    if format_letter.lower() == 'e':
        text_len += 4
    field_to_text_len[a_field] = max(text_len, len(field_to_field_label[a_field]))


def format_field_value(field, value):
    field = field.lower().strip()
    if field in field_to_multiplier.keys():
        value *= field_to_multiplier[field]
    number_str = value.__format__(field_to_format_strs[field])
    return number_str


def format_and_center(field, value):
    return format_field_value(field, value).center(field_to_text_len[field])


def calc_left_right_spaces(text, total_space_available):
    if len(text) > total_space_available:
        return '', '', ' ' * (len(text) - total_space_available)
    else:
        right_space = (total_space_available - len(text)) // 2
        left_space = total_space_available - right_space - len(text)
        return ' ' * left_space, ' ' * right_space, ''


class Res(NamedTuple):
    """
    base class for resonator parameters then a determined by a single fit
    Needs to be mixed in with a typing.NamedTuple
    """

    def __getitem__(self, item):
        if isinstance(item, str):
            return getattr(self, item)
        elif isinstance(item, int):
            return self._asdict()[self._fields[item]]
        elif isinstance(item, slice):
            return tuple([self._asdict()[field] for field in self._fields[item]])
        else:
            raise TypeError(f"ResFit indices must be str, int, or slice, not {type(item)}")

    def header_items(self):
        for field in self._fields:
            yield field

    def header(self):
        header_str = ''
        for field in self.header_items():
            header_str += f'{field},'
        return header_str[:-1]

    def input_items(self):
        for field in self.header_items():
            if field.lower() not in derived_params:
                yield field

    def __str__(self):
        return_str = ''
        for field in self.header_items():
            return_str += f"{self.__getattribute__(field)},"
        return return_str[:-1]

    def __call__(self, *args):
        if not args:
            args = self._fields
        values_list = []
        for arg in args:
            if arg in self._fields:
                value = self.__getattribute__(arg)
            else:
                value = None
            if value is None:
                value = float('nan')
            values_list.append(value)
        return tuple(values_list)

    def console(self, label: str = None, print_header: bool = True, fields=None):
        if label is None:
            label = ''
        if fields is None:
            found_fields = {}
            # non-derived parameters
            for field in self.input_items():
                value = self.__getattribute__(field)
                if value is not None:
                    found_fields[field.lower()] = value
            # derived parameters
            found_derived = {}
            Qi = self.Qi
            Qc = self.Qc
            if (Qc is None or Qi is None) and 'qr' in found_fields.keys() and 'amp' in found_fields.keys():
                Qc, Qi = calc_qc_qi(qr=self.Qr, amp=self.amp)
            if Qi is not None:
                found_derived['qi'] = Qi
            if Qc is not None:
                found_derived['qc'] = Qc
            if self.chi_sq is not None:
                found_derived['chi_sq'] = self.chi_sq
            if self.red_chi_sq is not None:
                found_derived['red_chi_sq'] = self.red_chi_sq
            if self.p_value is not None:
                found_derived['p_value'] = self.p_value
        else:
            found_fields = {}
            found_derived = {}
            for field in fields:
                if field.lower() in derived_params:
                    found_derived[field.lower()] = self.__getattribute__(field)
                else:
                    found_fields[field.lower()] = self.__getattribute__(field)
        # determine console output
        len_label = max(5, len(label))
        label_spaces = ' ' * len_label
        total_len_var_fit = sum([field_to_text_len[field] for field in found_fields.keys()])
        total_len_var_fit += len(found_fields.keys()) - 1
        var_fit_label = ' Variables fit '
        var_fit_left_space, var_fit_right_space, var_fit_extra_header = \
            calc_left_right_spaces(text=var_fit_label, total_space_available=total_len_var_fit)
        derived_label = ' Derived variables '
        total_len_derived = sum([field_to_text_len[field] for field in found_derived.keys()])
        total_len_derived += len(found_derived.keys()) - 1
        derived_fit_left_space, derived_fit_right_space, derived_fit_extra_header = \
            calc_left_right_spaces(text=derived_label, total_space_available=total_len_derived)
        # header only console output
        if print_header:
            if self.fr is None:
                fr_mhz = '(Frequency not found)'
            else:
                fr_mhz = f'{self.fr / 1e6:.2f} MHz'
            print(f'Resonator at {fr_mhz}')
            super_header = f'{label_spaces}|{var_fit_left_space}{var_fit_label}{var_fit_right_space}'
            sub_header = f'{label_spaces}|'
            first_field = True
            for field in found_fields.keys():
                field_len = field_to_text_len[field]
                formatted_field = field_to_field_label[field].center(field_len)
                if first_field:
                    sub_header += f'{formatted_field}'
                    first_field = False
                else:
                    sub_header += f'|{formatted_field}'
            sub_header += var_fit_extra_header
            if found_derived:
                super_header += '|' + derived_text(f'{derived_fit_left_space}{derived_label}{derived_fit_right_space}')
                sub_header += '|'
                first_field = True
                for field in found_derived.keys():
                    field_len = field_to_text_len[field]
                    formatted_field = field_to_field_label[field].center(field_len)
                    if first_field:
                        sub_header += derived_text(f'{formatted_field}')
                        first_field = False
                    else:
                        sub_header += derived_text(f'|{formatted_field}')

                if derived_fit_extra_header != '':
                    sub_header += derived_text(derived_fit_extra_header)
            print(super_header)
            print(sub_header)
        # values output for console
        values_str = label.__format__(f'<{len_label}')
        for field in found_fields.keys():
            values_str += '|' + format_and_center(field, found_fields[field])
        values_str += var_fit_extra_header
        if found_derived:
            first_field = True
            values_str += '|'
            for field in found_derived.keys():
                if first_field:
                    values_str += derived_text(format_and_center(field, found_derived[field]))
                    first_field = False
                else:
                    values_str += derived_text('|' + format_and_center(field, found_derived[field]))
            if derived_fit_extra_header != '':
                values_str += derived_text(derived_fit_extra_header)
        print(values_str)


class NonlinearIQResBase(NamedTuple):
    """ Base resonator parameters for a NonLinear IQ fit, nonlinear_iq_for_fitter()"""
    fr: Optional[float] = None
    Qr: Optional[float] = None
    amp: Optional[float] = None
    phi: Optional[float] = None
    a: Optional[float] = None
    i0: Optional[float] = None
    q0: Optional[float] = None
    tau: Optional[float] = None
    f0: Optional[float] = None
    chi_sq: Optional[float] = None
    p_value: Optional[float] = None
    red_chi_sq: Optional[float] = None
    Qi: Optional[float] = None
    Qc: Optional[float] = None


class NonlinearIQRes(NonlinearIQResBase, Res):
    """ The resonator parameters for a NonLinear IQ fit, nonlinear_iq_for_fitter()"""


class NonlinearMagResBase(NamedTuple):
    """ Base resonator parameters for a NonLinear Mag fit, fit_nonlinear_mag()"""
    fr: Optional[float] = None
    Qr: Optional[float] = None
    amp: Optional[float] = None
    phi: Optional[float] = None
    a: Optional[float] = None
    b0: Optional[float] = None
    b1: Optional[float] = None
    flin: Optional[float] = None
    chi_sq: Optional[float] = None
    p_value: Optional[float] = None
    red_chi_sq: Optional[float] = None
    Qi: Optional[float] = None
    Qc: Optional[float] = None


class NonlinearMagRes(NonlinearMagResBase, Res):
    """ The resonator parameters for a NonLinear Mag fit, fit_nonlinear_mag()"""


class LinearMagResBase(NamedTuple):
    """ Base resonator parameters for a Linear Mag fit, fit_linear_mag(), linear_mag()"""
    fr: Optional[float] = None
    Qr: Optional[float] = None
    amp: Optional[float] = None
    phi: Optional[float] = None
    b0: Optional[float] = None
    chi_sq: Optional[float] = None
    p_value: Optional[float] = None
    red_chi_sq: Optional[float] = None
    Qi: Optional[float] = None
    Qc: Optional[float] = None


class LinearMagRes(LinearMagResBase, Res):
    """ The resonator parameters for a Linear Mag fit, fit_linear_mag(), linear_mag()"""


# f_0, Q, Q_e_real, Q_e_imag, delay, phi_offset, fmin, A_mag, A_mag_slope
class ResonatorCableBase(NamedTuple):
    """ Base resonator parameters for a resonator_cable(), fit_resonator_cable()"""
    fr: Optional[float] = None
    Q: Optional[float] = None
    Q_e_real: Optional[float] = None
    Q_e_imag: Optional[float] = None
    delay: Optional[float] = None
    phi: Optional[float] = None
    f_min: Optional[float] = None
    A_mag: Optional[float] = None
    A_slope: Optional[float] = None
    chi_sq: Optional[float] = None
    p_value: Optional[float] = None
    red_chi_sq: Optional[float] = None
    Qi: Optional[float] = None
    Qc: Optional[float] = None


class ResonatorCable(ResonatorCableBase, Res):
    """ The resonator parameters for a resonator_cable(), fit_resonator_cable()"""


class Fit(NamedTuple):
    origin: str
    func: Callable
    guess: Optional[Res] = None
    result: Optional[Res] = None
    popt: Optional[np.ndarray] = None
    pcov: Optional[np.ndarray] = None
    f_data: Optional[np.ndarray] = None
    z_data: Optional[np.ndarray] = None
    mask: Optional[np.ndarray] = None
    flags: Optional[set] = None

    def __getitem__(self, item):
        if isinstance(item, str):
            # this could be a field of the result
            if item in self._fields:
                return getattr(self, item)
            elif item in self.result._fields:
                return getattr(self.result, item)
            elif '_err' in item.lower() and self.pcov is not None:
                field_name, _ = item.lower().rsplit('_err', 1)
                for field_index, field_name_check in list(enumerate(self._fields_results)):
                    if field_name_check.lower() == field_name:
                        return self.pcov[field_index, field_index]
                else:
                    raise KeyError(f'Error Key Field {item} not found in {self._fields_results}')

                return self._structured_array_results[item]
            elif item.lower() == 'fit':
                return self.popt, self.pcov
            elif item.lower() == 'fit_result':
                return self.z_fit()
            elif item.lower() == 'x0_result':
                return self.z_guess()
            elif item.lower() == 'x0':
                return self.guess
            elif item.lower() == 'z':
                return self.z_data
            else:
                raise KeyError(f'Unknown string key: {item}')
        elif isinstance(item, int):
            return self._asdict()[self._fields[item]]
        elif isinstance(item, slice):
            return tuple([self._asdict()[field] for field in self._fields[item]])
        else:
            raise TypeError(f"ResFit indices must be str, int, or slice, not {type(item)}")

    def __call__(self, *args):
        if not args:
            args = self._fields
        values_list = []
        for arg in args:
            if arg in self._fields:
                value = self.__getattribute__(arg)
            else:
                value = None
            if value is None:
                value = float('nan')
            values_list.append(value)
        return tuple(values_list)

    def __repr__(self):
        return f"Fit-{str(self.result).replace(',', '-')}"

    """ Base class for a resonator fit results """

    def z_fit(self) -> np.ndarray:
        """Return the complex impedance of the fit."""
        return self.func(self.f_data, *self.result[0:-len(derived_params)])

    def z_guess(self) -> np.ndarray:
        """Return the complex impedance of the guess."""
        return self.func(self.f_data, *self.guess[0:-len(derived_params)])

    def chi2_p(self) -> np.array:
        """Return the chi-squared and p-value of the fit."""
        # Chi-Square Goodness of Fit Test
        chi_square_test_statistic, p_value = stats.chisquare(f_obs=self.z_data, f_exp=self.z_fit())
        return chi_square_test_statistic, p_value

    def plot(self, show=True):
        plt.figure()
        z_guess = self.z_guess()
        z_fit = self.z_fit()
        if "mag" in self.origin:
            plt.plot(self.f_data / 10 ** 6, 20 * np.log10(np.abs(self.z_data)), 'o', label="Input Data")
            plt.plot(self.f_data / 10 ** 6, 20 * np.log10(np.abs(z_guess)), '-', label="Initial Guess")
            plt.plot(self.f_data / 10 ** 6, 20 * np.log10(np.abs(z_fit)), '-', label="Fit Results")
            plt.xlabel("Frequency (MHz)")
            plt.ylabel("Power (dB)")
            plt.legend()
        else:
            plt.axes().set_aspect('equal')
            plt.plot(np.real(self.z_data), np.imag(self.z_data), 'o', label="Input Data")
            plt.plot(np.real(z_guess), np.imag(z_guess), label="Initial Guess")
            plt.plot(np.real(z_fit), np.imag(z_fit), label="Fit Results")
            plt.xlabel('I')
            plt.ylabel('Q')
            plt.legend()
        if show:
            plt.show()

    def console(self, print_header=True, fields=None):
        if self.guess is None:
            if self.result is None:
                print('No fit results')
            else:
                self.result.console(label='Fit', print_header=print_header, fields=fields)
        else:
            self.guess.console(label='Guess', print_header=print_header, fields=fields)
            self.result.console(label='Fit', print_header=print_header, fields=fields)


class ResSet:
    def __init__(self, path: str = None, res_results: Sequence[Res] = None, res_fits: Sequence[Fit] = None,
                 verbose: bool = True):
        self.verbose = verbose
        if path is None:
            # use the default output directory, make it if it doesn't exist
            if not os.path.exists(abs_path_output_dir_default):
                os.mkdir(abs_path_output_dir_default)
            res_set_dir = os.path.join(abs_path_output_dir_default, 'res_set')
            if not os.path.exists(res_set_dir):
                os.mkdir(res_set_dir)
            # use a default file name buy do not overwrite an existing file
            count = 0
            base_path = os.path.join(res_set_dir, f'results{count:02}')
            path_test_csv = f'{base_path}.csv'
            path_test_toml = f'{base_path}.toml'
            # loop until we find a pair of file names that do not exist
            while os.path.exists(path_test_csv) and os.path.exists(path_test_toml):
                count += 1
                base_path = os.path.join(res_set_dir, f'results{count:02}')
                path_test_csv = f'{base_path}.csv'
                path_test_toml = f'{base_path}.toml'
            self.res_set_dir = res_set_dir
            self.path = base_path
        else:
            self.res_set_dir, basename = os.path.split(path)
            if '.' in basename:
                basename, ext = basename.rsplit('.', 1)

            # use the user specified path
            self.path = os.path.join(self.res_set_dir, basename)
            # make a new directory if it doesn't exist
            if not os.path.exists(self.res_set_dir):
                os.mkdir(self.res_set_dir)

        # set in self.read(), self.add_results(), self.add_res_fits methods
        self._results = set()
        # set only in the self.add_res_fits method
        self._fit_results = {}
        # set in self.order_and_validate() method
        self._fields_results = None
        self._list_results = None
        self._list_fit_results = None
        self._structured_array_results = None
        # removed indexes when a None results is encountered, # set in self.add_res_fits() and add_res_fits() methods
        self.removed_indexes = []

        if res_results is None and res_fits is None:
            if os.path.exists(f'{self.path}.csv') or os.path.exists(f'{self.path}.toml'):
                if self.verbose:
                    print(f'No fits on initialization, triggering a read in of the the results.')
                self.read()
            else:
                # null state setup
                pass
        else:
            if res_results is not None:
                self.add_results(res_results)
            if res_fits is not None:
                self.add_res_fits(res_fits)

    def __iter__(self):
        # iterate over the results in order of resonator frequency
        for result in sorted(self._results, key=attrgetter('fr')):
            yield result

    def __len__(self):
        return len(self._results)

    def __getitem__(self, item):
        if isinstance(item, int):
            return self._list_results[item]
        elif isinstance(item, slice):
            return self._list_results[item]
        elif isinstance(item, str):
            if item in self._fields_results:
                return self._structured_array_results[item]
            elif item in Fit._fields:
                return self._structured_array_fit_results[item]
            else:
                raise KeyError(f'Unknown string key: {item}')
        else:
            raise TypeError(f"ResSet indices must be str, int, or slice, not {type(item)}")

    def __call__(self, *args) -> dict:
        if not args:
            # when no args are specified
            args = list(self._fields_results) + list(Fit._fields)
        results_fields = []
        fits_fields = []
        for arg in args:
            if arg in self._fields_results:
                results_fields.append(arg)
            elif arg in Fit._fields:
                fits_fields.append(arg)
        results_dict = {arg: self._structured_array_results[arg] for arg in results_fields}
        if fits_fields:
            fits_dict = {arg: [] for arg in fits_fields}
            for fit in self._list_fit_results:
                for arg in fits_fields:
                    if fit is None:
                        fits_dict[arg].append(None)
                    else:
                        fits_dict[arg].append(fit.__getattribute__(arg))
            results_dict.update(fits_dict)
        return results_dict

    def __add__(self, other):
        if not isinstance(other, ResSet):
            raise TypeError(f"unsupported operand type(s) for +: 'ResSet' and '{type(other)}'")
        return ResSet(res_results=list(self._results) + list(other._results),
                      res_fits=list(self._fit_results.values()) + list(other._fit_results.values()),
                      verbose=self.verbose)

    def order_and_validate(self):
        # make list data objects
        self._list_results = list(self)
        # get the fields list from the first item
        self._fields_results = self._list_results[0]._fields
        # get the fit results list, which can have None values
        self._list_fit_results = []
        for result in self._list_results:
            if result in self._fit_results:
                self._list_fit_results.append(self._fit_results[result])
            else:
                self._list_fit_results.append(None)
        # make the numpy structured array data object
        values_results = [result() for result in self._list_results]
        columns_values_results = [np.array(column_values) for column_values in zip(*values_results)]
        dtypes_results = [(arg, str(column_array.dtype)) for arg, column_array in zip(self._fields_results,
                                                                                      columns_values_results)]
        self._structured_array_results = np.array(values_results, dtype=dtypes_results)

    def read(self, res_tuple=NonlinearIQRes):
        read_file_name_toml = f'{self.path}.toml'
        read_file_name_csv = f'{self.path}.csv'
        if os.path.exists(read_file_name_toml):
            self.read_toml(read_file_name=read_file_name_toml, res_tuple=res_tuple)
        elif os.path.exists(read_file_name_csv):
            self.read_csv(read_file_name=read_file_name_csv, res_tuple=res_tuple)
        else:
            raise FileNotFoundError(f'No results file found at {self.path} with .toml or .csv extension.')

    def read_toml(self, read_file_name, res_tuple=NonlinearIQRes):
        raise NotImplementedError('reading a toml method is not yet implemented.')
        list_fit_results = toml.load(read_file_name)
        for fit_result in list_fit_results:
            print('here')

    def read_csv(self, read_file_name, res_tuple=NonlinearIQRes):
        """Read in the results from a csv file."""
        if self.verbose:
            print(f'Reading results from file at: {filename_text(self.path)}')
        header_keys = []
        with open(read_file_name, 'r') as f:
            for line in f.readlines():
                if not header_keys:
                    # read the header, see what can be mapped into the NamedTuple specified with res_tuple
                    header_keys_raw = line.strip().split(',')
                    one_key_found = False
                    for header_key in header_keys_raw:
                        if header_key in res_tuple._fields:
                            header_keys.append(header_key)
                            one_key_found = True
                        else:
                            header_keys.append(None)
                    if not one_key_found:
                        raise ValueError(f'No valid header keys found in {self.path} ' +
                                         f'for the header keys {header_keys_raw} and tuple fields: {res_tuple._fields}')
                else:
                    # read the data after a successful header read
                    row_dict = {header_key: value_formatted for header_key, value_formatted
                                in zip(header_keys, line_format(line)) if header_key is not None}
                    self._results.add(res_tuple(**row_dict))
        self.order_and_validate()

    def write(self, csv: bool = True, toml: bool = False):
        if csv:
            self.write_csv()
        if toml:
            raise NotImplementedError('Writing to toml is not yet implemented.')
            self.write_toml()

    def write_csv(self):
        write_file_name = f'{self.path}.csv'
        with open(write_file_name, 'w') as f:
            first_row = True
            for result in self:
                # write the header
                if first_row:
                    f.write(result.header() + '\n')
                    first_row = False
                # write the results
                f.write(str(result) + '\n')
        if self.verbose:
            print(f'{write_text("Results Written")} to file at: {filename_text(self.path)}')

    def write_toml(self):
        write_file_name = f'{self.path}.toml'
        dict_of_fits = {repr(self._fit_results[result]): self._fit_results[result] for result in self._fit_results}
        with open(write_file_name, 'w') as f:
            toml.dump(dict_of_fits, f)
        if self.verbose:
            print(f'{write_text("Results Written")} to file at: {filename_text(self.path)}')

    def add_results(self, res_results: Sequence[Res]):
        for list_index, res_result in list(enumerate(res_results)):
            if res_result is None:
                self.removed_indexes.append(list_index)
            else:
                self._results.add(res_result)
        self.order_and_validate()

    def add_res_fits(self, res_fits: Sequence[Fit]):
        for list_index, res_fit in list(enumerate(res_fits)):
            if res_fit is None:
                self.removed_indexes.append(list_index)
            else:
                result = res_fit.result
                self._results.add(result)
                self._fit_results[result] = res_fit
        self.order_and_validate()

    def plot(self, global_flags=None, plot_title=None, plot_frames=True, flags_types=None):
        # plotter want frequencies and z values with shape n_frequency_point x n_res x n_sweeps
        # also for fitted parameters it wants a list fitted names for the legend
        # and an array that is n_res x len(fitted parameters)

        # Make the frequency and z arrays firs

        # Make the frequency and z arrays for original data and fitted
        n_iq_points = len(self._fit_results[next(iter(self))].z_data)  # assumes all data is the same size
        frequencies = np.zeros((n_iq_points, len(self)))
        fitted_frequencies = np.zeros((n_iq_points, len(self)))
        z_values = np.zeros((n_iq_points, len(self)), dtype='complex')
        fitted_z_values = np.zeros((n_iq_points, len(self)), dtype='complex')
        for i, result in enumerate(self):
            fit_result = self._fit_results[result]
            frequencies[:, i] = fit_result.f_data  # input frequencies nominally the same as res_freq_array
            fitted_frequencies[:, i] = fit_result.f_data  # input frequencies nominally the same as res_freq_array
            z_values[:, i] = fit_result.z_data
            fitted_z_values[:, i] = fit_result.z_fit()

        # stack the data with fit data
        multi_sweep_freqs = np.dstack((np.expand_dims(frequencies, axis=2), np.expand_dims(fitted_frequencies, axis=2)))
        multi_sweep_z = np.dstack((np.expand_dims(z_values, axis=2), np.expand_dims(fitted_z_values, axis=2)))

        # now get the fitted values
        # first get names of fitted values
        data_names = []
        formats = []
        result = next(iter(self))  # grab first fit for inspection
        for field in result._fields:
            if getattr(result, field) is not None:
                data_names.append(field)
                formats.append(field_to_field_label[field.lower()] + ': {:' + field_to_format_strs[field.lower()] + '}')

        fitted_parameters = np.zeros((len(self), len(data_names)))
        for i, result in enumerate(self):
            for j, field in enumerate(data_names): # this gets rid of the None
                value = getattr(result, field)
                if value is None or np.isnan(value):
                    value = -99.99
                if field in field_to_multiplier.keys():
                    fitted_parameters[i, j] = value * field_to_multiplier[field.lower()]
                else:
                    fitted_parameters[i, j] = value

        # format flags for the interactive plot
        if global_flags is None:
            flags = [list(fit.flags) for fit in self._list_fit_results]
        else:
            flags = [list(fit.flags.update(global_flags)) for fit in self._list_fit_results]
        # run the plotter
        ip = InteractivePlot(multi_sweep_freqs, multi_sweep_z, retune=False, combined_data=fitted_parameters,
                             combined_data_names=data_names, combined_data_format=formats,
                             sweep_labels=['Data', 'Fit'], flags=flags, flags_types=flags_types,
                             plot_title=plot_title, plot_frames=plot_frames, verbose=self.verbose)
        # flag and remove data based on the results of the interactive plot.
        removed_flag = False
        for result_index, (result, flags) in list(enumerate(zip(self, ip.flags))):
            fit = self._fit_results[result]
            if flags:
                fit.flags.update(flags)
            if result_index in ip.res_indexes_removed:
                fit.flags.add('remove')
                self._results.remove(result)
                del self._fit_results[result]
                removed_flag = True
        if removed_flag:
            self.order_and_validate()
        return ip



import numpy as np
import matplotlib.pyplot as plt

from submm.KIDs.res.fit_funcs import linear_mag


def brute_force_linear_mag_fit(f_hz, z, ranges, n_grid_points, error=None, plot=False):
    """
    Parameters
    ----------
    f_hz : numpy.array
        frequencies Hz
    z : numpy.array
        complex or abs of s21
    ranges : numpy.array
        The ranges for each parameter as in:
        np.asarray(([f_low,Qr_low,amp_low,phi_low,b0_low],[f_high,Qr_high,amp_high,phi_high,b0_high]))
    n_grid_points: int
        How finely to sample each parameter space. This can be very slow for n>10, an increase by a factor of 2 will
        take 2**5 times longer to marginalize over you must minimize over the unwanted axes of sum_dev
        i.e. for fr np.min(np.min(np.min(np.min(fit['sum_dev'],axis = 4),axis = 3),axis = 2),axis = 1)
    error : numpy.array, optional (default = None)
        The error on the complex or abs of s21, used for weighting squares
    plot : bool, optional (default = False)
        If true, will plot the fit and the data
    """
    if error is None:
        error = np.ones(len(f_hz))

    fs = np.linspace(ranges[0][0], ranges[1][0], n_grid_points)
    Qrs = np.linspace(ranges[0][1], ranges[1][1], n_grid_points)
    amps = np.linspace(ranges[0][2], ranges[1][2], n_grid_points)
    phis = np.linspace(ranges[0][3], ranges[1][3], n_grid_points)
    b0s = np.linspace(ranges[0][4], ranges[1][4], n_grid_points)
    evaluated_ranges = np.vstack((fs, Qrs, amps, phis, b0s))

    a, b, c, d, e = np.meshgrid(fs, Qrs, amps, phis, b0s, indexing="ij")  # always index ij

    evaluated = linear_mag(f_hz, a, b, c, d, e)
    data_values = np.reshape(np.abs(z) ** 2, (abs(z).shape[0], 1, 1, 1, 1, 1))
    error = np.reshape(error, (abs(z).shape[0], 1, 1, 1, 1, 1))
    sum_dev = np.sum(((np.sqrt(evaluated) - np.sqrt(data_values)) ** 2 / error ** 2),
                     axis=0)  # comparing in magnitude space rather than magnitude squared

    min_index = np.where(sum_dev == np.min(sum_dev))
    index1 = min_index[0][0]
    index2 = min_index[1][0]
    index3 = min_index[2][0]
    index4 = min_index[3][0]
    index5 = min_index[4][0]
    fit_values = np.asarray((fs[index1], Qrs[index2], amps[index3], phis[index4], b0s[index5]))
    fit_values_names = ('f0', 'Qr', 'amp', 'phi', 'b0')
    fit_result = linear_mag(f_hz, fs[index1], Qrs[index2], amps[index3], phis[index4], b0s[index5])

    marginalized_1d = np.zeros((5, n_grid_points))
    marginalized_1d[0, :] = np.min(np.min(np.min(np.min(sum_dev, axis=4), axis=3), axis=2), axis=1)
    marginalized_1d[1, :] = np.min(np.min(np.min(np.min(sum_dev, axis=4), axis=3), axis=2), axis=0)
    marginalized_1d[2, :] = np.min(np.min(np.min(np.min(sum_dev, axis=4), axis=3), axis=1), axis=0)
    marginalized_1d[3, :] = np.min(np.min(np.min(np.min(sum_dev, axis=4), axis=2), axis=1), axis=0)
    marginalized_1d[4, :] = np.min(np.min(np.min(np.min(sum_dev, axis=3), axis=2), axis=1), axis=0)

    marginalized_2d = np.zeros((5, 5, n_grid_points, n_grid_points))
    # 0 _
    # 1 x _
    # 2 x x _
    # 3 x x x _
    # 4 x x x x _
    #  0 1 2 3 4
    marginalized_2d[0, 1, :] = marginalized_2d[1, 0, :] = np.min(np.min(np.min(sum_dev, axis=4), axis=3), axis=2)
    marginalized_2d[2, 0, :] = marginalized_2d[0, 2, :] = np.min(np.min(np.min(sum_dev, axis=4), axis=3), axis=1)
    marginalized_2d[2, 1, :] = marginalized_2d[1, 2, :] = np.min(np.min(np.min(sum_dev, axis=4), axis=3), axis=0)
    marginalized_2d[3, 0, :] = marginalized_2d[0, 3, :] = np.min(np.min(np.min(sum_dev, axis=4), axis=2), axis=1)
    marginalized_2d[3, 1, :] = marginalized_2d[1, 3, :] = np.min(np.min(np.min(sum_dev, axis=4), axis=2), axis=0)
    marginalized_2d[3, 2, :] = marginalized_2d[2, 3, :] = np.min(np.min(np.min(sum_dev, axis=4), axis=1), axis=0)
    marginalized_2d[4, 0, :] = marginalized_2d[0, 4, :] = np.min(np.min(np.min(sum_dev, axis=3), axis=2), axis=1)
    marginalized_2d[4, 1, :] = marginalized_2d[1, 4, :] = np.min(np.min(np.min(sum_dev, axis=3), axis=2), axis=0)
    marginalized_2d[4, 2, :] = marginalized_2d[2, 4, :] = np.min(np.min(np.min(sum_dev, axis=3), axis=1), axis=0)
    marginalized_2d[4, 3, :] = marginalized_2d[3, 4, :] = np.min(np.min(np.min(sum_dev, axis=2), axis=1), axis=0)

    if plot:
        levels = [2.3, 4.61]  # delta chi squared two parameters 68 90 % confidence
        fig_fit = plt.figure(-1)
        axs = fig_fit.subplots(5, 5)
        for i in range(0, 5):  # y starting from top
            for j in range(0, 5):  # x starting from left
                if i > j:
                    # plt.subplot(5,5,i+1+5*j)
                    # axs[i, j].set_aspect('equal', 'box')
                    extent = [evaluated_ranges[j, 0], evaluated_ranges[j, n_grid_points - 1], evaluated_ranges[i, 0],
                              evaluated_ranges[i, n_grid_points - 1]]
                    axs[i, j].imshow(marginalized_2d[i, j, :] - np.min(sum_dev), extent=extent, origin='lower',
                                     cmap='jet')
                    axs[i, j].contour(evaluated_ranges[j], evaluated_ranges[i],
                                      marginalized_2d[i, j, :] - np.min(sum_dev), levels=levels, colors='white')
                    axs[i, j].set_ylim(evaluated_ranges[i, 0], evaluated_ranges[i, n_grid_points - 1])
                    axs[i, j].set_xlim(evaluated_ranges[j, 0], evaluated_ranges[j, n_grid_points - 1])
                    axs[i, j].set_aspect((evaluated_ranges[j, 0] - evaluated_ranges[j, n_grid_points - 1]) / (
                            evaluated_ranges[i, 0] - evaluated_ranges[i, n_grid_points - 1]))
                    if j == 0:
                        axs[i, j].set_ylabel(fit_values_names[i])
                    if i == 4:
                        axs[i, j].set_xlabel("\n" + fit_values_names[j])
                    if i < 4:
                        axs[i, j].get_xaxis().set_ticks([])
                    if j > 0:
                        axs[i, j].get_yaxis().set_ticks([])

                elif i < j:
                    fig_fit.delaxes(axs[i, j])

        for i in range(0, 5):
            # axes.subplot(5,5,i+1+5*i)
            axs[i, i].plot(evaluated_ranges[i, :], marginalized_1d[i, :] - np.min(sum_dev))
            axs[i, i].plot(evaluated_ranges[i, :], np.ones(len(evaluated_ranges[i, :])) * 1., color='k')
            axs[i, i].plot(evaluated_ranges[i, :], np.ones(len(evaluated_ranges[i, :])) * 2.7, color='k')
            axs[i, i].yaxis.set_label_position("right")
            axs[i, i].yaxis.tick_right()
            axs[i, i].xaxis.set_label_position("top")
            axs[i, i].xaxis.tick_top()
            axs[i, i].set_xlabel(fit_values_names[i])

        # axs[0,0].set_ylabel(fit_values_names[0])
        # axs[4,4].set_xlabel(fit_values_names[4])
        axs[4, 4].xaxis.set_label_position("bottom")
        axs[4, 4].xaxis.tick_bottom()

    # make a dictionary to return
    fit_dict = {'fit_values': fit_values, 'fit_values_names': fit_values_names, 'sum_dev': sum_dev,
                'fit_result': fit_result, 'marginalized_2d': marginalized_2d, 'marginalized_1d': marginalized_1d,
                'evaluated_ranges': evaluated_ranges}  # , 'x0':x0, 'z':z}
    return fit_dict



import pyvisa


class synthesizer(object):

    # Default constructor used to call the instrument at the given
    # GPIBaddress and initializes the object to the instrument
    def __init__(self):
        self.GPIBaddress = 'GPIB0::15::INSTR'
        rm = pyvisa.ResourceManager("@py")
        self.obj1 = rm.open_resource(self.GPIBaddress)

    # This method returns the CURRENT frequency that synthesizer is at
    # Using the built in command - OF1
    def get_frequency(self):  # does not work yet
        frequency = float(self.obj1.query('FREQ?'))
        return frequency

    # This method will set the frequency of the Anritsu to the
    # specified amount. The given freq_value needs to be within
    # the range of 10MHz - 20GHz.
    def set_frequency(self, freq_value):

        # If value is below/above a certain threshold,
        # let user know to re-enter

        if(freq_value < 10000000):
            print(freq_value)
            print("Invalid value - Needs to be greater than 10MHz")
        elif(freq_value > 20000000000):
            print("Invalid value - needs to be smaller than 20GHz")
        else:  # matlab code 'FREQ %.9f GHZ\n',fGHz
            cmd_str = 'FREQ %.9f GHZ' % (freq_value/1e9)
            self.obj1.write(cmd_str)
#freq = self.get_frequency()
# print(freq)

    # This method will return the CURRENT power that the frequency
    # is on
    def get_power(self):
        power = float(self.obj1.query('POW:LEV?'))
        return power

    # This method will set the power to the specified amount
    def set_power(self, power_value):  # matlab code POW:LEV %f dBM\n'

        # If the value is below/above a certain threshold,
        # let user know to re-enter
        if(power_value < -20):
            print("Invalid value - Needs to be greater than -20dbm")
        elif(power_value > 30):
            print("Invalid value - Needs to be less than 30dbm")
        else:
            cmd_str = 'POW:LEV %f dBM' % power_value
            self.obj1.write(cmd_str)

    # This method will turn the output to ON 'OUTP:STAT?'
    def turn_output_on(self):
        self.obj1.write('OUTP:STAT ON')

    # This method will turn the output to OFF
    def turn_output_off(self):
        self.obj1.write('OUTP:STAT OFF')

    def get_output(self):
        output = bool(int(self.obj1.query('OUTP:STAT?')))
        return output

    # This method will print out to terminal all of the available
    # functions that are included in this class
    def help(self):
        print("")
        print("\t Available Functions")
        print("\t --------------------")
        print("\t set_frequency(Hz)")
        print("\t get_frequency()")
        print("\t get_power()")
        print("\t set_power(Dbm)")
        print("\t turn_output_on()")
        print("\t turn_output_off()")
        print("\t get_output()")



import os
from setuptools import setup, find_packages


def get_file_paths(local_dir, desired_file_types):
    data_files = []
    # loop over  all the sample data files and get  with the appropriate extension
    for file_basename in os.listdir(local_dir):
        try:
            filename_prefix, filename_extension = file_basename.rsplit('.', 1)
        except ValueError:
            # ignore files without an extension
            pass
        else:
            # do this if no ValueError was raised
            if filename_extension in desired_file_types:
                data_files.append(os.path.join(local_dir, file_basename))
    return data_files


# Example data files
desired_example_file_types = {'csv', 'pkl', 'txt', 'xlsx', 'xls', 'fits', 'mat'}
local_samples_dir = os.path.join('submm', 'sample_data')
sample_data_files = get_file_paths(local_dir=local_samples_dir, desired_file_types=desired_example_file_types)


# Demo files
desired_demo_file_types = {'ipynb'}
local_demos_dir = os.path.join('submm', 'demo')
demo_files = get_file_paths(local_dir=local_demos_dir, desired_file_types=desired_demo_file_types)


setup(name='submm',
      version='0.3.1',
      description='Python routines for submm astronomy instrumentation',
      author='Jordan Wheeler',
      author_email='wheeler1711@gmail.com',
      packages=find_packages(),
      url="https://github.com/Wheeler1711/submm_python_routines",
      data_files=[(local_samples_dir, sample_data_files),
                  (local_demos_dir, demo_files)],
      include_package_data=True,
      python_requires='>3.7',
      install_requires=['numpy<1.23,>=1.18',
                        'matplotlib>=3.5.2',
                        'scipy',
                        'numba',
                        'tqdm',
                        'importlib-metadata',
                        'toml']
      )



import numpy as np
import matplotlib.pyplot as plt
from matplotlib.backends.backend_pdf import PdfPages
from scipy import fftpack
import nitime.algorithms as tsa #pip install nitime


#code to do principal component analysis on correlated detectors.
#I find that when you use PCA every component you subtract off loses most of its signal.
# so I guess to make gains in S/N you need to have more S/N gain from the PCA then you lose from losing the pixel
# for a spectrometer you probably don't want to lose important pixels so you should pick pixels with no lines of interest


def PCA(arr,ncomp,verbose = True,sample_rate = 488.28125,filename = "",**keywords):

    pdf_pages = PdfPages('PCA_diag'+filename+'.pdf') # name of the pdf produced

    # do mean subration
    mean_vec_arr  = np.mean(arr,axis = 0)
    arr = arr - mean_vec_arr

    # normalize by the variance
    var = np.std(arr,axis = 0)**2
    arr = arr/var

    covmat = np.dot(arr.T,arr)#/data_arr.shape[0]

    #compute the eigvectors and eigenvalues
    evals,evects = np.linalg.eig(covmat)

    #sort the eigenvectors and values by largest to smallest
    idx = evals.argsort()[::-1]   
    evals = evals[idx]
    evects = evects[:,idx]

    #make the eigenfunction array
    efuncarr = np.dot(arr,evects)
    # pick components to subtract
    efuncarr[:,0:ncomp] = 0
    res_index_used = idx[0:ncomp]
    #transform back
    arr_clean = np.inner(efuncarr,evects)
    
    if verbose:
        print( "the covariance matrix is")
        print( covmat)
        print("")
        print("the eigenvalues are")
        print(evals)
        print("")
        print("the eigenvectors are")
        print(evects)
        print("")
    fig = plt.figure(1,figsize = (12,6))
    plt.subplot(1,2,1)
    plt.title("Eigenvalues")
    plt.plot(np.log10(np.abs(evals)),label = "eigenvalues")
    plt.plot(np.log10(np.abs(evals[0:ncomp])),'o',label = "components removed")
    plt.xlabel("Eigenvalue index")
    plt.ylabel("Eigenvalue Magnitude")
    plt.legend()


    plt.subplot(1,2,2)
    plt.title("Log10(covariance matrix)")
    plt.imshow(np.log10(covmat),interpolation = 'nearest')
    plt.colorbar()

    pdf_pages.savefig(fig)
    plt.close()

    #need to unnormalize by the varaince
    arr = arr*var
    arr_clean = arr_clean*var

    #need to add the mean back in to get the right units
    arr = arr + mean_vec_arr
    arr_clean = arr_clean + mean_vec_arr
    
    #block for plotting the correlation between the two most corrrelated detectors    
    fig = plt.figure(2,figsize = (12,6))
    plt.title("Cross correlation of two most correlated detectors")
    # make a figure that show the correlation between the two most correlated detectors
    cov_matrix_2 = covmat
    for i in range(0,cov_matrix_2.shape[0]):
        cov_matrix_2[i,i] = 0 #zeros out diagonal obviously detecotrs are correlated with them selves
    max_corr_index1 = np.where(cov_matrix_2 == np.max(cov_matrix_2))[0][0]
    max_corr_index2 = np.where(cov_matrix_2 == np.max(cov_matrix_2))[0][1]

    # this is kind of a neat package for doing cross correlation on time streams
    # see this page for explanation
    # http://nbviewer.jupyter.org/github/mattijn/pynotebook/blob/master/ipynotebooks/Python2.7/2016/2016-05-25%20cross-spectral%20analysis.ipynb
    #calculate the cross correlation
    f, pcsd_est = tsa.multi_taper_csd(np.vstack((arr[:,max_corr_index1],arr[:,max_corr_index2])), Fs=sample_rate, low_bias=True, adaptive=False, sides='onesided')
    fki = pcsd_est.diagonal().T[0]
    fkj = pcsd_est.diagonal().T[1]
    cij = pcsd_est.diagonal(+1).T.ravel()
    #calculate the coherene
    coh = np.abs(cij)**2 / (fki * fkj) 

    
    plt.loglog(f,fki,label = "Resonator psd index = "+str(max_corr_index1))
    plt.loglog(f,fkj,label = "Resonator psd index = "+str(max_corr_index2))
    ylim = plt.ylim()
    plt.loglog(f,np.abs(cij),label = "Cross psd")
    plt.ylim(ylim[0],ylim[1])
    plt.legend()

    pdf_pages.savefig(fig)
    plt.close()

    fig = plt.figure(3,figsize = (12,6))
    plt.semilogx(f,np.abs(coh))
    plt.title("Coherence")
    plt.xlabel("Frequency Hz")
    plt.ylim(0,1)

    pdf_pages.savefig(fig)
    plt.close()


    pdf_pages.close()


    PCA_dict = {'arr':arr,
                    'cleaned':arr_clean,
                    'ncomp':ncomp,
                    'res_index_used':res_index_used,
                    'cov_matrix':covmat}
    return PCA_dict



#!/usr/bin/env python

"""This is an implementation to the model for disordered superconductivity using
Dynes model of elastic scattering DOS broadening. This is mostly following the
Zemlicka et al. 2015 derivation of a Dynes type model with modified propagators
in the Nam equations, making the calculations very similar to the AG model in
the 'disordered_model.py' file

Joseph Redford 10/21/2021
"""

import multiprocessing as mp

import numpy as np
from scipy.integrate import quad
from scipy.optimize import root_scalar
from scipy import interpolate

#constants
k_B = 1.380649e-23 #J/K
h = 6.62607e-34 #J*s

def calc_delta_int(Delta0, gamma, kT):
    def integrand(E, Delta0, gamma, delta, kT):
        term1 = np.real(np.power(E * E - (Delta0 - 1.j * gamma)**2, -0.5))
        term2 = np.tanh(0.5 * E / kT) * np.real(np.power(E * E - (delta
                - 1.j * gamma)**2, -0.5))
        return term1 - term2

    def prepare_integrand(delta):
        int_fn = lambda E: integrand(E, Delta0, gamma, delta, kT)
        return int_fn
    root_fn = lambda delta: quad(prepare_integrand(delta), 0., np.inf)[0]
    delta_sol = root_scalar(root_fn, x0=Delta0, x1 = 0.)

    return delta_sol.root


def propagator_n(E, delta, gamma):
    denom = np.power(((E + 1.j * gamma)**2.) - delta * delta, -0.5) 
    num = np.sign(E) * (E + 1.j * gamma)

    return num * denom


def propagator_p(E, delta, gamma):
    denom = np.power(((E + 1.j * gamma)**2.) - delta * delta, -0.5) 
    num = np.sign(E) * delta

    return num * denom


def calc_sigma2_thermal(kT, delta, h_nu, gamma):
    def integrand(E):
        if kT > 0.:
            beta = 1. / kT
        else:
            beta = np.inf
        occupation = np.tanh(0.5 * beta * (E + h_nu))
        prop_term = np.imag(propagator_n(E, delta, gamma)) * np.real(
                propagator_n(E + h_nu, delta, gamma)) + np.imag(
                propagator_p(E, delta, gamma)) * np.real(
                propagator_p(E + h_nu, delta, gamma))
        return occupation * prop_term

    int_val, _ = quad(integrand, -np.inf, np.inf)
    return  int_val/ h_nu


def calc_sigma1_thermal(kT, delta, h_nu, gamma):
    def integrand(E):
        if kT > 0.:
            beta = 1. / kT
        else:
            beta = np.inf
        occupation = 0.5 * (np.tanh(0.5 * beta * (E + h_nu))
                - np.tanh(0.5 * beta * E))
        prop_term = np.real(propagator_n(E, delta, gamma)) * np.real(
                propagator_n(E + h_nu, delta, gamma)) + np.real(
                propagator_p(E, delta, gamma)) * np.real(
                propagator_p(E + h_nu, delta, gamma))
        return occupation * prop_term

    int_val, _ = quad(integrand, -np.inf, np.inf)
    return  int_val/ h_nu



def sigma2_fit_fn(kT, Delta0, h_nu, gamma):
    current_delta = calc_delta_int(Delta0, gamma, kT)
    sigma2_ratio = calc_sigma2_thermal(kT, current_delta, h_nu, gamma)

    return sigma2_ratio


def thermal_Qi(kT, Delta0, h_nu, gamma):
    current_delta = calc_delta_int(Delta0, gamma, kT)
    sigma2_ratio = calc_sigma2_thermal(kT, current_delta, h_nu, gamma)
    sigma1_ratio = calc_sigma1_thermal(kT, current_delta, h_nu, gamma)
    
    return sigma2_ratio / sigma1_ratio


def n_qp(kT, delta, gamma, N_0):
    def integrand(E):
        pop_level = 1. - np.tanh(0.5 * E / kT)
        DOS = np.real((E + 1.j * gamma) * np.power(
                (E + 1.j * gamma)**2. - delta * delta, -0.5))

        return pop_level * DOS

    N_pop, _ = quad(integrand, 0., np.inf)

    return 2. * N_0 * N_pop


def S2_calc(kT, delta, gamma, N_0, h_nu):
    n_thermal = n_qp(T, delta, gamma, N_0)
    sigma2_0 = calc_sigma2_thermal(0., delta, h_nu, gamma)
    S2_prefix = 0.5 * (calc_sigma2_thermal(kT, delta, h_nu, gamma)
            - sigma2_0)/ (sigma2_0 * n_thermal)

    return S2_prefix


def GR_Sxx(kT, delta, gamma, N_0, h_nu, tau, tau_max, V):
    #Delta = calc_delta_int(T_c, alpha, T, N_V)
    #Delta_0 = calc_delta_int(T_c, alpha, 0., N_V)
    n_thermal = n_qp(kT, delta, gamma, N_0)
    if n_thermal == 0.:
        return 0.
    sigma2_0 = calc_sigma2_thermal(0., delta, h_nu, gamma)
    S2_prefix = (sigma2_0 - calc_sigma2_thermal(kT, delta, h_nu, gamma)) \
            / (sigma2_0 * n_thermal)
    fluctuation_term = 2. * tau * (1. + tau / tau_max) * n_thermal / V

    return S2_prefix * S2_prefix * fluctuation_term




#!/usr/bin/env python

"""This is an implementation to the model for disordered TiN mostly following
the approach laid out in Coumou's thesis, but broken up and reparameterized in
ways I suspect will be more conducive to numerical calculation and integration.
It ends up being an implementation of the Abrikosov-Gor'kov (AG) model with the
admittance calculations of the film calculated with Nam generalizations of the
Matthis-Bardeen integrals.

Joseph Redford 5/31/2021
"""

import multiprocessing as mp

import numpy as np
from scipy.integrate import quad
from scipy.optimize import root_scalar
from scipy import interpolate

#constants
k_B = 1.380649e-23 #J/K
h = 6.62607e-34 #J*s


def get_usadel_x(E, Delta, eta):
    """This calculates the value of e^(i \\theta) = x from the Usadel equation.
    I think this is better than fully backing out theta or cos/sin theta because
    this then becomes a problem of finding polynomial roots and all later
    quantities can be put into terms of x with algebra and the Euler identity

    E is the energy level being calculated
    Delta is the average gap energy of the disordered superconductor
    eta is the depairing energy/gap broadening parameter

    In terms of units, E and Delta should just have to share units but could
    be put in any energy unit or even a temperature using a kT scaling
    eta is dimensionless
    """
    poly_coeff = [0.5j * Delta * eta, Delta + E, 0., Delta - E,
            -0.5j * Delta * eta]
    roots = np.roots(poly_coeff)
    roots = roots[np.argsort(np.imag(roots))]
    #the two roots that remain purely imaginary and are not physical are
    #maximum and minimum in the imaginary component for positive energy values
    #once the real component is non-zero the two correct roots should only
    #differ by sign of the real component, so taking positive real part to
    #consistently be on the same branch
    roots = roots[1:3]
    delta = roots[1] - roots[0]
    if np.imag(delta) > np.real(delta):
        return roots[0]
    else:
        return roots[1]


def get_usadel_x2(E, Delta, alpha):
    """This calculates the value of e^(i \\theta) = x from the Usadel equation.
    I think this is better than fully backing out theta or cos/sin theta because
    this then becomes a problem of finding polynomial roots and all later
    quantities can be put into terms of x with algebra and the Euler identity

    E is the energy level being calculated
    Delta is the average gap energy of the disordered superconductor
    alpha is the gap broadening parameter

    In terms of units, E and Delta should just have to share units but could
    be put in any energy unit or even a temperature using a kT scaling
    eta is dimensionless
    """
    poly_coeff = [1.j * alpha, 2. * (Delta + E), 0., 2. * (Delta - E),
            -1.j * alpha]
    roots = np.roots(poly_coeff)
    roots = roots[np.argsort(np.imag(roots))]
    #the two roots that remain purely imaginary and are not physical are
    #maximum and minimum in the imaginary component for positive energy values
    #once the real component is non-zero the two correct roots should only
    #differ by sign of the real component, so taking positive real part to
    #consistently be on the same branch
    roots = roots[1:3]
    delta = roots[1] - roots[0]
    if np.imag(delta) > np.real(delta):
        return roots[0]
    else:
        return roots[1]


def g1(E1, E2, Delta, alpha):
    """This is the g1 greens function used in the dissipation calculations 

    E1 and E2 are the two energy arguments of the greens function
    Delta is the average gap energy
    alpha is the the depairing energy/gap broadening parameter

    just like the usadel function, E1, E2, and Delta could be any units as
    long as they are the same
    """
    x1 = get_usadel_x2(np.abs(E1), Delta, alpha)
    x2 = get_usadel_x2(np.abs(E2), Delta, alpha)
    x1_sq_mag = np.real(x1 * np.conj(x1)) #I am not sure the real is necessary,
            #but ensures that this is not complex data type
    x2_sq_mag = np.real(x2 * np.conj(x2))
    
    g1_val = np.real(x1 * x1_sq_mag + x1) * np.real(x2 * x2_sq_mag + x2)
    g1_val += np.real(x1 * x1_sq_mag - x1) * np.real(x2 * x2_sq_mag - x2)
    g1_val /= (x1_sq_mag * x2_sq_mag)

    return 0.25 * g1_val


def g2(E1, E2, Delta, alpha):
    """This is the g2 greens function used in the inductance calculations 

    E1 and E2 are the two energy arguments of the greens function
    Delta is the average gap energy
    eta is the the depairing energy/gap broadening parameter

    just like the usadel function, E1, E2, and Delta could be any units as
    long as they are the same
    """
    x1 = get_usadel_x2(np.abs(E1), Delta, alpha)
    x2 = get_usadel_x2(np.abs(E2), Delta, alpha)
    x1_sq_mag = np.real(x1 * np.conj(x1)) #I am not sure the real is necessary,
            #but ensures that this is not complex
    x2_sq_mag = np.real(x2 * np.conj(x2))

    g2_val = 0.5 * (x1_sq_mag * x2_sq_mag - 1) / (x1_sq_mag * x2_sq_mag)
    g2_val *= np.imag(x1) * np.real(x2)

    return 0.25 * g2_val #double check the 0.25 prefix


def get_Eg_fraction(eta, cutoff, precision):
    """Since only relative values of E and Delta to each other matter,
    I am just going to find the fraction of Delta Eg is at (the effective
    fractional reduction of the gap due to eta). So eta is and cutoff is the
    only parameters

    This search will be one sided as the real value of x is 0 (to within
        precision of the root finder routine) until suddenly you cross Eg

    I will use cutoff to set a threshold above the noise of the root finder
        imprecision, that generates some small real parts to some of the sub-gap
        roots and then a "binary search" type stepping and testing. If the step
        is above the Eg step again, else halve the step. It is primitive, but
        should work just fine.
    """
    current = 1.
    step = 0.5
    while step >= precision:
        if current - step < 0.:
            step *= 0.5 #shouldn't go negative, not sure the Usadel solver
                    #will work, and the physical system should be symmetric
                    #around 0 anyway
            continue
        temp_x = get_usadel_x(current-step, 1., eta)
        if np.real(temp_x) > cutoff:
            current -= step
        else:
            step *= 0.5

    return current


def analytical_Eg(Delta, alpha):
    if Delta < alpha:
        return 0.
    else:
        return (Delta**(2./3.) - alpha**(2./3.))**1.5 / Delta


def thermal_gap_suppression(T, Delta_0):
    Tc = Delta_0 / 1.76 #assuming delta is really k_B Delta, so as always, can
            #keep this unit agnostic
    eff_delta = Delta_0 * np.sqrt(1. - T / Tc) #this is just a first test to see
        #if the gap suppression is necessary to get my results to match
        #the plots of Coumou, I remember this being a simple approximation
        #of the gap suppression with temperature

    #return eff_delta
    return Delta_0


def fermi(E, mu, beta):
    """I use $\\beta = 1 / (k T)$ just because then I can have E, mu, and beta
    in similar units and this can accept anything that is consistent in the unit
    system.

    E is the energy level
    mu is the chemical potential
    beta is inverse kT so should be in inverse units of E and mu
    """
    ex_term = np.exp(beta * (E - mu))

    return 1. / (1. + ex_term)


def calc_sigma1_thermal(T, Delta, h_nu, alpha):
    """This calculates the $\\frac{\\sigma_1}{\\sigma_n}$ with numerical
    integration of the greens function formalism from Nam.
    As with everything I will assume the proper factors of k and h were applied
    such that T, Delta, and h_nu are all in same units of energy/effective
    temperature.

    T is the temperature
    Delta is the average gap energy
    h_nu is the microwave photon energy
    eta is the depairing energy/gap broadening parameter
    """
    E_g = analytical_Eg(Delta, alpha)
    beta = 1. / T
    integrand1 = lambda E: g1(E, E + h_nu, Delta, alpha) * np.tanh(
            0.5 * beta * (E + h_nu))
    integrand2 = lambda E: g1(E, E + h_nu, Delta, alpha) * (
            np.tanh(0.5 * beta * (E + h_nu)) - np.tanh(0.5 * E * beta))
    term1, _ = quad(integrand1, E_g - h_nu, -E_g) #fine tune integration
            #parameters later
    term2, _ = quad(integrand2, E_g, np.inf)

    return (term1 + term2) / h_nu


def calc_sigma2_thermal(T, Delta, h_nu, alpha):
    """This calculates the $\\frac{\\sigma_1}{\\sigma_n}$ with numerical
    integration of the greens function formalism from Nam.
    As with everything I will assume the proper factors of k and h were applied
    such that T, Delta, and h_nu are all in same units of energy/effective
    temperature.

    T is the temperature
    Delta is the average gap energy
    h_nu is the microwave photon energy
    eta is the depairing energy/gap broadening parameter
    """
    #eff_delta = thermal_gap_suppression(T, Delta)
    E_g = analytical_Eg(Delta, alpha)
    #E_g *= eff_delta #convert in our delta for limits of integration
    if T <= 0.:
        fermi_val = lambda E: 0.
    else:
        beta = 1. / T
        fermi_val = lambda E : fermi(E, 0., beta)
    integrand1 = lambda E: g2(E, E + h_nu, Delta, alpha) * (1 - 2 * fermi_val(
                E + h_nu))
    integrand2 = lambda E: g2(E + h_nu, E, Delta, alpha) * (1 - 2 * fermi_val(
                E))
    limit = np.max([E_g - h_nu, -E_g]) #for KIDS I don't think we will ever
        #end up with -E_g being the greater
    term1, _ = quad(integrand1, limit, np.inf) #fine tune integration
            #parameters later
    term2, _ = quad(integrand2, E_g, np.inf)

    return (term1 + term2) / h_nu


def iterative_gap(T, T_c, alpha):
    def f(series_cutoff, delta, alpha, T, T_c):
        series_diff = np.inf
        series_sum = 0.
        i = 0
        while series_diff > series_cutoff:
            omega = (2. * i + 1.) * T * np.pi
            #x = matsubara_usadel_x(omega, curr_delta, alpha)
            #sin_term = -0.5j * (x + 1. / x)
            #term = (curr_delta / omega - np.real(sin_term))
            term = iterative_gap_matsubara_arg(omega, delta, alpha)
            series_sum += term
            series_diff = np.abs(term)
            i += 1
        new_delta = np.real(series_sum * 2. * np.pi * T)
        return new_delta

    #alpha = T_c * eta / 1.76
    cutoff = 1e-6
    min_T = 0.002
    if T < min_T:
        T = min_T
        #return Delta_0
    series_cutoff = 5e-8
    #I will assume T is actually kT in same units as Delta_0
    #using secant method for root finding
    curr_delta = T_c * 3.
    #if alpha / T_c > 0.1:
        #last_delta = (1. - (alpha / T_c)**0.6) * 1.76 * T_c
    #else:
    last_delta = 0.9 * 1.76 * T_c
    curr_f =  f(series_cutoff, curr_delta, alpha, T, T_c)
    last_f =  f(series_cutoff, last_delta, alpha, T, T_c)
    diff = np.abs(curr_delta - last_delta)
    a = np.log(T_c / T)
    while diff > cutoff:
        #secant iteration
        new_delta = (last_delta * (curr_f - a * curr_delta) - curr_delta * (
                last_f - a * last_delta)) / (curr_f - a * curr_delta - last_f
                + a * last_delta)
        diff = np.abs(curr_delta - new_delta)
        #print(diff)
        last_delta = curr_delta
        curr_delta = new_delta
        last_f = curr_f
        curr_f = f(series_cutoff, curr_delta, alpha, T, T_c)

    return curr_delta


def matsubara_usadel_x(omega, delta, alpha):
    poly_coeff = [alpha, 2. * (omega - 1.j * delta), 0.,
            -2. * (omega + 1.j * delta), -alpha]
    roots = np.roots(poly_coeff)
    roots = sorted(roots)

    return roots[3]


def iterative_gap_matsubara_arg(omega, delta, alpha):
    y = delta / omega
    poly_coeff = [alpha * alpha,
            2. * alpha * delta - 4. * alpha * alpha * y,
            6. * alpha * alpha * y * y - 6. * alpha * delta * y + omega * omega
            - alpha * alpha + delta * delta,
            6. * alpha * delta * y * y - 4. * alpha * alpha * y * y * y
            - 2. * (omega * omega - alpha * alpha + delta * delta) * y
            - 2. * alpha * delta,
            alpha * alpha * (y**4.) - 2. * alpha * delta * (y**3.)
            + (omega * omega - alpha * alpha + delta * delta) * y * y
            + (2. * alpha / omega - 1.) * delta * delta]
    roots = np.roots(poly_coeff)
    #roots = roots[np.argsort(-1.j * roots)] #sort by imaginary
    #roots = np.sort(roots)

    return np.real(roots[3]) #should be real root, making sure return data type
            #is not complex, but float

def calc_sigma2_Tc(T, T_c, h_nu, alpha):
    delta = iterative_gap(T, T_c, alpha)
    if delta < 1e-8:
        return 0.

    return calc_sigma2_thermal(T, delta, h_nu, alpha)


def calc_sigma1_Tc(T, T_c, h_nu, alpha):
    delta = iterative_gap(T, T_c, alpha)
    if delta < 1e-7:
        return 1.

    return calc_sigma1_thermal(T, delta, h_nu, alpha)


def calc_eff_Tc(T_c0, alpha):
    delta_thres = 1e-5
    zero_thres = 1e-8
    if alpha == 0.:
        return T_c0
    #right now, brute force binary search
    low_delta = iterative_gap(0., T_c0, alpha)
    low_Tc = 0.
    high_Tc = T_c0
    high_delta = iterative_gap(high_Tc, T_c0, alpha)
    while (high_Tc - low_Tc) / high_Tc >= 1e-5:
        next_Tc = 0.5 * (high_Tc + low_Tc)
        print(next_Tc)
        next_delta = iterative_gap(next_Tc, T_c0, alpha)
        if next_delta > zero_thres:
            low_Tc = next_Tc
            low_delta = next_delta
        else:
            high_Tc = next_Tc
            high_delta = next_delta

    return new_Tc


def map_fn(vals):
    ret_val = calc_eff_Tc(vals[0], vals[1])
    print(vals[1])

    return ret_val


def calc_Tc_var(Tc, delta_alpha):
    alpha_vals = np.linspace(0.01, Tc, int(np.ceil(Tc /delta_alpha)))
    #eff_Tc = np.ndarray(alpha_vals.shape)
    pool = mp.Pool()
    vals = [(Tc, alpha) for alpha in alpha_vals]
    eff_Tc = map(map_fn, vals)
    eff_Tc = [val for val in eff_Tc]
    print(eff_Tc)
    eff_Tc = np.array(eff_Tc)
    #for i, alpha in enumerate(alpha_vals):
    #    print(i)
    #    eff_Tc[i] = calc_eff_Tc(Tc, alpha)

    return eff_Tc, alpha_vals



def calc_delta_temp(delta, T_c, alpha):
    def f(series_cutoff, delta, alpha, T, T_c):
        #if T > T_c:
        #    return T
        #if T <= 0.:
        #    return T_c - T
        series_diff = np.inf
        series_sum = 0.
        i = 0
        while series_diff > series_cutoff:
            omega = (2. * i + 1.) * T * np.pi
            term = iterative_gap_matsubara_arg(omega, delta, alpha)
            series_sum += term
            series_diff = np.abs(term)
            i += 1
        new_delta = np.real(series_sum * 2. * np.pi * T)
        return new_delta

    #alpha = T_c * eta / 1.76
    cutoff = 1e-6
        #return Delta_0
    series_cutoff = 5e-8
    #I will assume T/T_c is actually kT in same units as Delta_0
    #using secant method for root finding
    current_T = T_c
    last_T = 1.1 * T_c
    root_func = lambda T_val: f(series_cutoff, delta, alpha, T_val, T_c) \
            - np.log(T_c / T_val) * delta
    current_T = root_scalar(root_func, x0=last_T, x1=current_T, maxiter=int(1e3))

    return current_T.root


def iterative_gap2(T, T_c, alpha):
    def f(series_cutoff, delta, alpha, T, T_c):
        series_diff = np.inf
        series_sum = 0.
        i = 0
        while series_diff > series_cutoff:
            omega = (2. * i + 1.) * T * np.pi
            #x = matsubara_usadel_x(omega, curr_delta, alpha)
            #sin_term = -0.5j * (x + 1. / x)
            #term = (curr_delta / omega - np.real(sin_term))
            term = iterative_gap_matsubara_arg(omega, delta, alpha)
            series_sum += term
            series_diff = np.abs(term)
            i += 1
        new_delta = np.real(series_sum * 2. * np.pi * T)
        return new_delta

    #alpha = T_c * eta / 1.76
    cutoff = 1e-6
    series_cutoff = 5e-8
    min_T = 0.02 * T_c
    if T < min_T:
        T = min_T
        #return Delta_0
    curr_delta = T_c * 1.76
    last_delta = 2.1 * T_c
    a = np.log(T / T_c)
    root_fn = lambda delta: f(series_cutoff, delta, alpha, T, T_c) \
            + a * delta
    curr_delta = root_scalar(root_fn, x0=last_delta, x1=curr_delta, bracket=(1e-3, 2 * 1.76 * T_c), maxiter=int(1e3))

    return curr_delta.root


def calc_delta_int(T_c, alpha, T, N_V):
    def sine_theta(E, delta):
        x = get_usadel_x2(E, delta, alpha)
        return (x * x - 1 ) / (2.j * x)
    debeye_temp = 0.5 * 1.76 * T_c * np.exp(1. / N_V)
    if T / T_c < 1e-2:
        temp_depend = lambda E: 0.
    else:
        beta = 1. / T
        temp_depend = lambda E: fermi(E, 0., beta)
    def disorder_root(delta):
        integrand = lambda E: np.imag(sine_theta(E, delta)) * (1. -
                2. * temp_depend(E))
        integral_val, _ = quad(integrand, 0., debeye_temp)
        return delta - N_V * integral_val
    delta_0 = root_scalar(disorder_root, x0=1.76*T_c, x1=1.6 * T_c, maxiter=int(1e4))

    return delta_0.root


def calc_NV(T_c, alpha):
    def root_fn(NV):
        if NV < 1e-4:
            debeye_temp = np.inf
        else:
            debeye_temp = 0.5 * 1.76 * T_c * np.exp(1. / NV)
        delta_0 = 1.76 * T_c
        f = lambda E: E / np.sqrt(E * E + delta_0 * delta_0)
        term1, _ = quad(f, 0., debeye_temp, ) #fine tune integration
        return delta_0 - NV * term1

    N_V = root_scalar(root_fn, x0=0.1, x1=0.2, maxiter=int(1e4))
    N_V = N_V.root

    return N_V


def calc_T_from_delta_int(T_c, alpha, delta, N_V):
    def sine_theta(E, delta):
        x = get_usadel_x2(E, delta, alpha)
        return (x * x - 1 ) / (2.j * x)
    debeye_temp = 0.5 * 1.76 * T_c * np.exp(1. / N_V)
    def disorder_root(T):
        if T / T_c < 1e-2:
            temp_depend = lambda E: 0.
        else:
            beta = 1. / T
            temp_depend = lambda E: fermi(E, 0., beta)
        integrand = lambda E: np.imag(sine_theta(E, delta)) * (1. -
                2. * temp_depend(E))
        integral_val, _ = quad(integrand, 0., debeye_temp)
        return delta - N_V * integral_val
    T_sol = root_scalar(disorder_root, x0=0., x1=0.1 * T_c, maxiter=int(1e4))

    return T_sol.root


def calc_real_Tc(T_c_orig, alpha, res):
    NV = calc_NV(T_c_orig, alpha)
    delta_0 = calc_delta_int(T_c_orig, alpha, 0., NV)
    last_d = 0.1 * delta_0
    last_T = calc_T_from_delta_int(T_c_orig, alpha, last_d, NV)
    curr_d = 0.05 * delta_0
    curr_T = calc_T_from_delta_int(T_c_orig, alpha, curr_d, NV)
    estimate = curr_T + curr_d * (curr_T - last_T) / (last_d - curr_d)
    while estimate - curr_T > res:
        print(estimate)
        last_d = curr_d
        last_T = curr_T
        curr_d = 0.2 * last_d
        curr_T = calc_T_from_delta_int(T_c_orig, alpha, curr_d, NV)
        estimate = curr_T + curr_d * (curr_T - last_T) / (last_d - curr_d)
    return estimate


def invert_eff_Tc(alpha_start, alpha_stop, delta):
    N_steps = int(np.ceil((alpha_stop - alpha_start) / delta))
    alpha_vals = np.linspace(alpha_start, alpha_stop, N_steps)
    results = np.ndarray(alpha_vals.shape)
    delta_0 = np.ndarray(alpha_vals.shape)
    for i, a in enumerate(alpha_vals):
        NV = calc_NV(1., a)
        delta_0[i] = calc_delta_int(1., a, 0., NV)
        results[i] = calc_real_Tc(1., a, 1e-5)
    print(delta_0)
    invert = interpolate.interp1d(alpha_vals / delta_0, results)
    alpha_revert = interpolate.interp1d(alpha_vals / delta_0, alpha_vals)
    def inversion(T_c_eff, alpha_eff):
        alpha_ratio = alpha_eff / T_c_eff
        T_c = T_c_eff / invert(alpha_ratio)
        alpha = alpha_revert(alpha_ratio) * T_c
        return T_c, alpha

    return inversion


def map_fn(args):
    return args[0], args[1], calc_delta_int(1., args[2], args[3], args[4])


def precalc_temp_delta(T_start, T_stop, alpha_start, alpha_stop, step_size):
    N_steps_a = int(np.ceil((alpha_stop -  alpha_start) / step_size))
    N_steps_T = int(np.ceil((T_stop - T_start) / step_size))
    T_vals = np.linspace(T_start, T_stop, N_steps_T)
    alpha_vals = np.linspace(alpha_start, alpha_stop, N_steps_a)
    results = np.ndarray((N_steps_a, N_steps_T))
    args = []
    for i, alpha in enumerate(alpha_vals):
        NV = calc_NV(1., alpha)
        print(i)
        for j, T in enumerate(T_vals):
            args.append((i, j, alpha, T, NV))
            #results[i][j] = calc_delta_int(1., alpha, T, NV)
    pool = mp.Pool()
    map_results = pool.map(map_fn, args)
    for data in map_results:
        results[data[0]][data[1]] = data[2]
    return interpolate.interp2d(T_vals, alpha_vals, results)


def generate_sigma2_fn(T_start, T_stop, alpha_start, alpha_stop, step_size):
    inverter = invert_eff_Tc(alpha_start, alpha_stop, step_size)
    delta_fn = precalc_temp_delta(T_start, T_stop, alpha_start, alpha_stop,
            step_size)
    def ret_fn(T, T_c, h_nu, alpha):
        real_Tc, real_alpha = inverter(T_c, alpha)
        delta = real_Tc * delta_fn(T / real_Tc, real_alpha / real_Tc)[0]
        return calc_sigma2_thermal(T, delta, h_nu, real_alpha)

    return ret_fn


def calc_Qi(T_vals, T_c, alpha, h_nu):
    inverter = invert_eff_Tc(0.001, 0.9, 1e-2)
    real_Tc, real_alpha = inverter(T_c, alpha / T_c)
    NV = calc_NV(real_Tc, real_alpha)
    results = np.ndarray(len(T_vals))
    for i, T in enumerate(T_vals):
        delta = calc_delta_int(real_Tc, real_alpha, T, NV)
        results[i] = np.abs(calc_sigma2_thermal(T, delta, h_nu, alpha
                ) / calc_sigma1_thermal(T, delta, h_nu, alpha))
        #results[i] = calc_sigma1_thermal(T, delta, h_nu, alpha)

    return results


def N_qp(T, Delta, alpha, N_0):
    def integrand(E):
        x_val = get_usadel_x2(E, Delta, alpha)
        x_sq_mag = np.real(x_val * np.conj(x_val))
        cos_val = np.real(x_val * x_sq_mag + x_val) / x_sq_mag
        pop_level = 1. - np.tanh(0.5 * E / T)
        return cos_val * pop_level

    E_g = analytical_Eg(Delta, alpha)
    N_int, _ = quad(integrand, E_g, np.inf)

    return N_0 * N_int


def S2_calc(T, Delta, alpha, N_0, h_nu):
    n_thermal = N_qp(T, Delta, alpha, N_0)
    S2_prefix = 0.5 * calc_sigma2_thermal(T, Delta, h_nu, alpha) / (
            calc_sigma2_thermal(0., Delta, h_nu, alpha) * n_thermal)

    return S2_prefix


def GR_Sxx(T, Delta, alpha, N_0, h_nu, tau, tau_max, V):
    #Delta = calc_delta_int(T_c, alpha, T, N_V)
    #Delta_0 = calc_delta_int(T_c, alpha, 0., N_V)
    n_thermal = N_qp(T, Delta, alpha, N_0)
    if n_thermal == 0.:
        return 0.
    sigma2_0 = calc_sigma2_thermal(0., Delta, h_nu, alpha)
    S2_prefix = 0.5 * (calc_sigma2_thermal(T, Delta, h_nu, alpha) - sigma2_0) \
            / (sigma2_0 * n_thermal)
    fluctuation_term = 4. * tau * (1. + tau / tau_max) * n_thermal / V

    return S2_prefix * S2_prefix * fluctuation_term




import gc
import time
import copy
import platform

import numpy as np
from scipy import interpolate
import matplotlib.pyplot as plt
from matplotlib.path import Path
from matplotlib.widgets import LassoSelector
from matplotlib.backends.backend_pdf import PdfPages
import tqdm

from submm.KIDs.res.utils import colorize_text, text_color_matplotlib, autoscale_from_data

'''
Tools for handling resonator iq sweeps 
i.e. finding minimum, maximum of iq sweep arrays
plotting iq sweep arrays
retuning resonators
finding correct readout power level
modified from https://github.com/sbg2133/kidPy/
'''


def find_max_didq(z, look_around):
    Is = np.real(z)
    Qs = np.imag(z)
    pos_offset_I = np.roll(Is, 1, axis=0)
    neg_offset_I = np.roll(Is, -1, axis=0)
    pos_offset_Q = np.roll(Qs, 1, axis=0)
    neg_offset_Q = np.roll(Qs, -1, axis=0)
    pos_dist = np.sqrt((Is - pos_offset_I) ** 2 + (Qs - pos_offset_Q) ** 2)
    neg_dist = np.sqrt((Is - neg_offset_I) ** 2 + (Qs - neg_offset_Q) ** 2)
    ave_dist = (pos_dist + neg_dist) / 2.
    # zero out the last and first values
    ave_dist[0, :] = 0
    ave_dist[ave_dist.shape[0] - 1, :] = 0
    min_index = np.argmax(ave_dist[Is.shape[0] // 2 - look_around:Is.shape[0] // 2 + look_around],
                          axis=0) + (Is.shape[0] // 2 - look_around)
    return min_index


class SelectFromCollection:
    """
    Select indices from a matplotlib collection using `LassoSelector`.

    Selected indices are saved in the `ind` attribute. This tool fades out the
    points that are not part of the selection (i.e., reduces their alpha
    values). If your collection has alpha < 1, this tool will permanently
    alter the alpha values.

    Note that this tool selects collection objects based on their *origins*
    (i.e., `offsets`).

    Parameters
    ----------
    ax : `~matplotlib.axes.Axes`
        Axes to interact with.
    collection : `matplotlib.collections.Collection` subclass
        Collection you want to select from.
    alpha_other : 0 <= float <= 1
        To highlight a selection, this tool sets all selected points to an
        alpha value of 1 and non-selected points to *alpha_other*.
    """

    def __init__(self, ax, collection, alpha_other=0.3):
        self.canvas = ax.figure.canvas
        self.collection = collection
        self.alpha_other = alpha_other

        # set in self.update()
        self.xys = None
        self.update()
        self.Npts = len(self.xys)

        # Ensure that we have separate colors for each object
        self.fc = collection.get_facecolors()
        if len(self.fc) == 0:
            raise ValueError('Collection must have a facecolor')
        elif len(self.fc) == 1:
            self.fc = np.tile(self.fc, (self.Npts, 1))

        self.lasso = LassoSelector(ax, onselect=self.onselect)
        self.ind = []

    def update(self):
        self.xys = self.collection.get_offsets()

    def onselect(self, verts):
        path = Path(verts)
        self.ind = np.nonzero(path.contains_points(self.xys))[0]
        self.fc[:, -1] = self.alpha_other
        self.fc[self.ind, -1] = 1
        self.collection.set_facecolors(self.fc)
        self.canvas.draw_idle()

    def plot_reset(self):
        self.fc[:, -1] = 1
        self.collection.set_facecolors(self.fc)

    def disconnect(self):
        self.lasso.disconnect_events()
        self.plot_reset()
        self.canvas.draw_idle()


class InteractivePlot(object):
    """
    interactive plot for plot many resonators iq data
    chan freqs and z should have dimension n_iq points by n_res
    also can provide multiple sweeps for plotting by adding extra dimension
    i.e. chan freqs and z could have dimension n_iq points by n_res by n_sweeps
    combined data should have dimension n_res by n_different_types of data
    """

    log_y_data_types = {'chi_sq'}
    key_font_size = 9

    flags_types_default = ["collision", "shallow", 'no-res', 'remove', 'other']

    def __init__(self, chan_freqs, z, look_around=2, stream_data=None, retune=True, find_min=True,
                 combined_data=None, combined_data_names=None, sweep_labels=None, sweep_line_styles=None,
                 combined_data_format=None, flags=None, flags_types=None, plot_title=None, plot_frames=True,
                 verbose=True):
        if len(z.shape) < 3:  # only one sweep
            self.z = z.reshape((z.shape[0], z.shape[1], 1))
            self.chan_freqs = chan_freqs.reshape((chan_freqs.shape[0], chan_freqs.shape[1], 1))
        else:
            self.z = z
            self.chan_freqs = chan_freqs

        plt.rcParams['keymap.fullscreen'] = ['shift+=']  # remove ('f', 'ctrl+f'), make +

        self.Is = np.real(self.z)
        self.Qs = np.imag(self.z)
        self.find_min = find_min
        self.retune = retune
        self.combined_data = combined_data
        self.combined_data_format = combined_data_format
        self.combined_data_names = combined_data_names
        self.stream_data = stream_data
        self.targ_size = chan_freqs.shape[0]
        self.look_around = look_around
        self.plot_index = 0
        self.combined_data_index = 0
        self.res_index_override = np.asarray((), dtype=np.int16)
        self.override_freq_index = np.asarray((), dtype=np.int16)
        self.shift_is_held = False
        self.lasso_mode = False
        self.lasso_start_time = 0.0
        self.selector = None
        self.update_min_index()
        if flags is None:
            self.flags = []
            for i in range(chan_freqs.shape[1]):
                self.flags.append([])
        else:
            self.flags = flags
        if flags_types is None:
            self.flags_types = self.flags_types_default
        else:
            self.flags_types = flags_types
        self.verbose = verbose
        self.all_flags = set(self.flags_types)
        self.flag_type_index = 0
        if retune:
            self.combined_data_names = ['min index']
        # data remove variables for the interactive plot
        self.res_indexes_removed = set()
        self.res_indexes_staged = {}
        self.combined_x_width_over_y_height = 1.0

        # set up plot
        top = 0.94
        bottom = 0.05
        left = 0.08
        right = 0.99
        x_space = 0.075
        y_space = 0.09
        x_width_total = right - left - x_space
        y_height_total = top - bottom - y_space
        if self.combined_data is None:
            self.fig = plt.figure(1, figsize=(13, 6))
            # the magnitude plot - upper left quadrant
            mag_x_width = 0.5
            mag_y_height = y_height_total
            mag_figure_coords = [left, top - mag_y_height, mag_x_width, mag_y_height]
            # the IQ plot - upper right quadrant
            iq_x_width = x_width_total - mag_x_width
            iq_y_height = mag_y_height
            iq_figure_coords = [left + mag_x_width + x_space, top - iq_y_height, iq_x_width, iq_y_height]
            combined_figure_coords = None
            key_figure_coords = None
        else:
            self.fig = plt.figure(1, figsize=(13, 10))
            # the magnitude plot - upper left quadrant
            mag_x_width = 0.5
            mag_y_height = 0.45
            mag_figure_coords = [left, top - mag_y_height, mag_x_width, mag_y_height]
            # the IQ plot - upper right quadrant
            iq_x_width = x_width_total - mag_x_width
            iq_y_height = mag_y_height
            iq_figure_coords = [left + mag_x_width + x_space, top - iq_y_height, iq_x_width, iq_y_height]
            # the Key Figure area for the interactive plot instructions-key
            key_x_width = 0.2
            key_y_height = y_height_total - mag_y_height
            combined_x_width = right - left - key_x_width
            key_figure_coords = [left + combined_x_width, bottom, key_x_width, key_y_height]
            # the combined data plot - lower plane
            combined_y_height = key_y_height
            self.combined_x_width_over_y_height = combined_x_width / combined_y_height
            combined_figure_coords = [left, bottom, combined_x_width, combined_y_height]

        if plot_title is not None:
            self.fig.suptitle(plot_title, y=0.99)
        self.ax_mag = self.fig.add_axes(mag_figure_coords, frameon=plot_frames)
        self.ax_mag.set_ylabel("Power (dB)")
        self.ax_mag.set_xlabel("Frequency (MHz)")
        self.ax_iq = self.fig.add_axes(iq_figure_coords, frameon=plot_frames)
        self.ax_iq.set_ylabel("Q")
        self.ax_iq.set_xlabel("I")
        if combined_figure_coords is None:
            self.ax_combined = None
        else:
            self.ax_combined = self.fig.add_axes(combined_figure_coords, frameon=False, autoscale_on=False)
            self.ax_combined.set_ylabel("")
            self.ax_combined.set_xlabel("Resonator index")
            self.ax_key = self.fig.add_axes(key_figure_coords, frameon=False)
            self.ax_key.tick_params(axis='y', which='both', left=False, right=False, labelleft=False)
            self.ax_key.tick_params(axis='x', which='both', bottom=False, top=False, labelbottom=False)
            self.plot_instructions()
        self.fig.canvas.mpl_connect('key_press_event', self.on_key_press)
        self.fig.canvas.mpl_connect('key_release_event', self.on_key_release)
        self.fig.canvas.mpl_connect('button_press_event', self.onClick)
        if self.stream_data is not None:
            self.s2, = self.ax_iq.plot(np.real(self.stream_data[:, self.plot_index]),
                                       np.imag(self.stream_data[:, self.plot_index]), '.')

        if not sweep_line_styles:
            sweep_line_styles = ["-o"]
            for i in range(1, self.z.shape[2]):
                sweep_line_styles.append("-")
        if not sweep_labels:
            sweep_labels = ["Data 1"]
            for i in range(1, self.z.shape[2]):
                sweep_labels.append("Data " + str(i + 1))
        self._mag_lines = [self.ax_mag.plot(self.chan_freqs[:, self.plot_index, i] / 10 ** 6, 10 * np.log10(
            self.Is[:, self.plot_index, i] ** 2 + self.Qs[:, self.plot_index, i] ** 2), sweep_line_styles[i], mec="k",
                                            label=sweep_labels[i])
                           for i in range(0, self.z.shape[2])]

        self.ax_leg = self.ax_mag.legend()

        self._iq_lines = [self.ax_iq.plot(
            self.Is[:, self.plot_index, i], self.Qs[:, self.plot_index, i], sweep_line_styles[i], mec="k",
            label=sweep_labels[i])
            for i in range(0, self.z.shape[2])]

        if self.retune:
            self.p1, = self.ax_mag.plot(self.chan_freqs[self.min_index[self.plot_index], self.plot_index, 0] / 10 ** 6,
                                        10 * np.log10(
                                            self.Is[self.min_index[self.plot_index], self.plot_index, 0] ** 2 +
                                            self.Qs[self.min_index[self.plot_index], self.plot_index, 0] ** 2),
                                        '*', markersize=15, color='darkorchid')
            self.p2, = self.ax_iq.plot(self.Is[self.min_index[self.plot_index], self.plot_index, 0],
                                       self.Qs[self.min_index[self.plot_index], self.plot_index, 0], '*', markersize=15)

        center_freq_MHz = self.chan_freqs[self.chan_freqs.shape[0] // 2, self.plot_index, 0] / 10 ** 6
        self.ax_mag.set_title(f'{"%3.3f" % center_freq_MHz} MHz - Resonator Index: {self.plot_index:03}')
        if self.retune:
            self.ax_iq.set_title("Look Around Points " + str(self.look_around))
        # Say 'Hello' to the User
        if self.verbose:
            print("\nInteractive Resonance Plotting Activated")
        self.print_instructions()

        # combined plot variables used in the first initialization
        self.combined_data_points = None
        self.combined_data_highlight = None
        self.combined_data_crosshair_x = None
        self.combined_data_crosshair_y = None
        self.combined_staged_for_removal = None
        self.combined_staged_flagged = None
        self.combined_data_legend = None
        self.res_indexes = None
        self.combined_data_values = None
        self.combined_values_this_index = None
        self.pop_up_text = None
        if self.combined_data is None:
            self.res_indexes_original = np.arange(0, self.chan_freqs.shape[1])
            self.res_indexes = self.res_indexes_original
        else:
            # combined data initial formatting
            self.combined_data = np.asarray(self.combined_data)
            if len(self.combined_data.shape) == 1:
                self.combined_data = np.expand_dims(self.combined_data, 1)

            if not self.combined_data_format:
                self.combined_data_format = []
                for i in range(0, self.combined_data.shape[1]):
                    self.combined_data_format.append(self.combined_data_names[i] + ': {:g}')
            else:
                self.combined_data_format = self.combined_data_format
            self.res_indexes_original = np.arange(0, self.combined_data.shape[0])
            # run the initialization script
            self.combined_plot(ax_combined=self.ax_combined)

        plt.show(block=True)

    def combined_plot(self, ax_combined):
        if self.combined_data_points is not None:
            self.combined_data_points.remove()
            self.combined_data_points = None
        if self.combined_data_highlight is not None:
            self.combined_data_highlight.remove()
            self.combined_data_highlight = None
        if self.combined_data_crosshair_x is not None:
            self.combined_data_crosshair_x.remove()
            self.combined_data_crosshair_x = None
        if self.combined_data_crosshair_y is not None:
            self.combined_data_crosshair_y.remove()
            self.combined_data_crosshair_y = None
        gc.collect()
        # initialize or re-initialize the combined data for the plot
        if self.res_indexes_removed:
            removed_array = np.array(sorted(self.res_indexes_removed))
            self.res_indexes = np.delete(arr=self.res_indexes_original, obj=removed_array, axis=0)
            self.combined_data_values = np.delete(arr=self.combined_data, obj=removed_array, axis=0)
        else:
            self.res_indexes = self.res_indexes_original
            self.combined_data_values = self.combined_data
        # rest the plot index to the first resonator
        self.plot_index = self.res_indexes[0]
        data_name = self.combined_data_names[self.combined_data_index]
        # start setting the plot
        ax_combined.set_title(data_name)
        ax_combined.set_ylabel(data_name)
        if data_name in self.log_y_data_types:
            ax_combined.set_yscale('log')
        else:
            ax_combined.set_yscale('linear')
        # plot the points and define the curves handle to update the data later
        self.combined_values_this_index = self.combined_data_values[:, self.combined_data_index]
        color_array = []
        edge_color_array = []
        for res_index, flags_this_res in list(enumerate(self.flags)):
            if res_index not in self.res_indexes_removed:
                if flags_this_res:
                    color_array.append('chartreuse')
                    edge_color_array.append('navy')
                else:
                    color_array.append('black')
                    edge_color_array.append('darkorchid')

        self.combined_data_points = ax_combined.scatter(x=self.res_indexes, y=self.combined_values_this_index, s=60,
                                                        color=color_array, marker='o', edgecolors=edge_color_array)
        # highlighting and cross-hairs for the selected data point
        highlighted_data_value = self.combined_data[self.plot_index, self.combined_data_index]
        label = self.combined_data_format[self.combined_data_index].format(highlighted_data_value)
        x_pos = self.plot_index
        y_pos = highlighted_data_value
        # the highlight symbol
        self.combined_data_highlight, = ax_combined.plot(x_pos, y_pos, 'o', markerfacecolor="None",
                                                         markeredgecolor='darkorange', markersize=14, label=label)
        # x crosshair
        self.combined_data_crosshair_x = ax_combined.axvline(x=x_pos, color='firebrick', ls='-', linewidth=1)
        # y crosshair
        self.combined_data_crosshair_y = ax_combined.axhline(y=y_pos, color='firebrick', ls='-', linewidth=1)
        self.combined_data_legend = ax_combined.legend()
        # Staged for Removal - In 'Remove Mode' these point have an 'X' to denote that they are staged for removal.
        self.plot_staged_for_removal(ax_combined=ax_combined)
        # we only rescale manually (intentionally) for this axis
        ax_combined.autoscale()

    def get_stage_plot_points(self):
        staged_indexes = sorted(self.res_indexes_staged.keys())
        staged_values = [self.combined_data[staged_index, self.combined_data_index]
                         for staged_index in staged_indexes]
        removed_indexes = []
        flag_indexes = []
        removed_values = []
        flag_values = []
        for staged_index, staged_value in zip(staged_indexes, staged_values):
            flags = self.res_indexes_staged[staged_index]
            if 'remove' in flags or 'no-res' in flags:
                removed_indexes.append(staged_index)
                removed_values.append(staged_value)
            else:
                flag_indexes.append(staged_index)
                flag_values.append(staged_value)
        return removed_indexes, removed_values, flag_indexes, flag_values

    def plot_staged_for_removal(self, ax_combined):
        if self.combined_staged_for_removal is not None:
            self.combined_staged_for_removal.remove()
            self.combined_staged_for_removal = None
        if self.combined_staged_flagged is not None:
            self.combined_staged_flagged.remove()
            self.combined_staged_flagged = None
        if self.res_indexes_staged:
            removed_indexes, removed_values, flag_indexes, flag_values = self.get_stage_plot_points()
            if removed_indexes:
                self.combined_staged_for_removal, = ax_combined.plot(removed_indexes, removed_values, 'x', ls='None',
                                                                     markersize=12, color='firebrick')
            if flag_indexes:
                self.combined_staged_flagged, = ax_combined.plot(flag_indexes, flag_values, '1', ls='None',
                                                                 markersize=12, color='deepskyblue')

    def refresh_plot(self, autoscale=True):
        if len(self.flags[self.plot_index]) > 0:
            self.ax_mag.set_facecolor('lightyellow')
            self.ax_iq.set_facecolor('lightyellow')
        else:
            self.ax_mag.set_facecolor("None")
            self.ax_iq.set_facecolor("None")
        for i, mag_line in enumerate(self._mag_lines):
            mag_line[0].set_data(self.chan_freqs[:, self.plot_index, i] / 10 ** 6, 10 * np.log10(
                self.Is[:, self.plot_index, i] ** 2 + self.Qs[:, self.plot_index, i] ** 2))
        if self.retune:
            self.p1.set_data(self.chan_freqs[self.min_index[self.plot_index], self.plot_index, 0] / 10 ** 6,
                             10 * np.log10(self.Is[self.min_index[self.plot_index], self.plot_index, 0] ** 2 +
                                           self.Qs[self.min_index[self.plot_index], self.plot_index, 0] ** 2))

        self.ax_mag.relim()
        self.ax_mag.autoscale()
        center_freq_MHz = self.chan_freqs[self.chan_freqs.shape[0] // 2, self.plot_index, 0] / 10 ** 6
        self.ax_mag.set_title(f'{"%3.3f" % center_freq_MHz} MHz - Resonator Index: {self.plot_index:03}')
        if self.retune:
            self.ax_iq.set_title("Look Around Points " + str(self.look_around))
        for i, iq_line in enumerate(self._iq_lines):
            iq_line[0].set_data((self.Is[:, self.plot_index, i],
                                 self.Qs[:, self.plot_index, i]))
        if self.retune:
            self.p2.set_data(self.Is[self.min_index[self.plot_index], self.plot_index, 0],
                             self.Qs[self.min_index[self.plot_index], self.plot_index, 0])

        if self.stream_data is not None:
            self.s2.set_data(np.real(self.stream_data[:, self.plot_index]),
                             np.imag(self.stream_data[:, self.plot_index]))
        self.ax_iq.relim()
        self.ax_iq.autoscale()
        if self.combined_data is not None:
            data_type = self.combined_data_names[self.combined_data_index]
            self.ax_combined.set_title(data_type, color='black')
            self.ax_combined.set_ylabel(data_type)
            if data_type in self.log_y_data_types:
                self.ax_combined.set_yscale('log')
            else:
                self.ax_combined.set_yscale('linear')
            # reset the combined plot data
            self.combined_values_this_index = self.combined_data_values[:, self.combined_data_index]
            new_offsets = np.column_stack((self.res_indexes, self.combined_values_this_index))
            self.combined_data_points.set_offsets(new_offsets)
            # label for the value of the highlighted data point
            highlighted_value = self.combined_data[self.plot_index, self.combined_data_index]
            label = self.combined_data_format[self.combined_data_index].format(highlighted_value)
            x_pos = self.plot_index
            y_pos = highlighted_value
            self.combined_data_highlight.set_data(x_pos, y_pos)
            self.combined_data_legend.texts[0].set_text(label)
            if self.res_indexes_staged:
                removed_indexes, removed_values, flag_indexes, flag_values = self.get_stage_plot_points()
                if removed_indexes:
                    self.combined_staged_for_removal.set_data(removed_indexes, removed_values)
                if flag_indexes:
                    self.combined_staged_flagged.set_data(flag_indexes, flag_values)
            if autoscale:
                self.ax_combined.set_xlim(autoscale_from_data(self.res_indexes))
                plot_min, plot_max = autoscale_from_data(self.combined_values_this_index,
                                                         log_scale=data_type in self.log_y_data_types)
                self.ax_combined.set_ylim((plot_min, plot_max))
            self.combined_data_crosshair_x.set_xdata(x_pos)
            self.combined_data_crosshair_y.set_ydata(y_pos)
        # lasso tool
        if self.lasso_mode:
            self.selector.update()
        # Pop-up text window
        if self.pop_up_text is not None:
            self.pop_up_text.remove()
            self.pop_up_text = None
        if self.lasso_mode and time.time() < self.lasso_start_time + 15:
            self.popup_lasso()
        elif self.flags[self.plot_index]:
            pop_up_text = f"Res {self.plot_index} Flagged: {sorted(self.flags[self.plot_index])}"
            self.pop_up_text = self.ax_combined.text(0.5, 0.8, pop_up_text,
                                                     transform=self.ax_combined.transAxes, ha="center", va="center",
                                                     size=16, color="black", family="monospace",
                                                     bbox=dict(facecolor='Red', alpha=0.6))
        plt.draw()

    def update_min_index(self):
        if self.find_min:
            self.min_index = np.argmin(
                self.Is[self.targ_size // 2 - self.look_around:self.targ_size // 2 + self.look_around, :, 0] ** 2 +
                self.Qs[self.targ_size // 2 - self.look_around:self.targ_size // 2 + self.look_around, :, 0] ** 2,
                axis=0) + (self.targ_size // 2 - self.look_around)
        else:
            self.min_index = find_max_didq(self.z[:, :, 0], self.look_around)
        # handel overridden points
        for i, override_index in enumerate(self.res_index_override):
            self.min_index[override_index] = self.override_freq_index[i]

        if self.retune:
            self.combined_data = np.expand_dims(self.min_index, 1)

    def popup_lasso(self):
        self.pop_up_text = self.ax_combined.text(0.5, 0.8, "Press enter to accept selection",
                                                 transform=self.ax_combined.transAxes, ha="center", va="center",
                                                 size=16, color="yellow", family="monospace",
                                                 bbox=dict(facecolor='black', alpha=0.6))

    def instructions(self):
        instructions = [("left-arrow", "change resonator left", 'green'),
                        ("right-arrow", "change resonator right", 'cyan')]
        if self.retune:
            instructions.extend([('down-arrow', 'change look around points down', 'yellow'),
                                 ('up-arrow', 'change look around points up', 'blue')])
            if platform.system() == 'Darwin':
                instructions.extend([("Hold any letter a key and right click on the magnitude plot",
                                      "to override tone position", 'black')])
            else:
                instructions.extend([("Hold 'shift' and right click on the magnitude plot",
                                      "to override tone position", 'black')])
        if self.combined_data is not None:
            flag_type = self.flags_types[self.flag_type_index]
            instructions.extend([('down-arrow', 'change y-data type', 'yellow'),
                                 ('up-arrow', 'change y-data type', 'blue'),
                                 ('double-click', 'go to the resonator index', 'black')])
            if self.lasso_mode:
                instructions.extend([('Enter-Key', f'Stage Lassoed, flag: {flag_type}', 'red'),
                                     ('B-Key', "Exit Lasso-selection", 'white'),
                                     ('Click+Hold', "Draw to Lasso a group", 'black')])
            else:
                instructions.extend([('F-key', f'stage for flag: {flag_type}', 'red'),
                                     ('Z-Key', f'un-stage for flag: {flag_type}', 'cyan'),
                                     ('B-Key', "Lasso-select to stage", 'white'),
                                     ('D-key', 'change flag mode', 'yellow'),
                                     ('T-Key', 'commit all staged flagging', 'blue'),
                                     ('E-Key', 'clear all staged flagging', 'green')])
            if self.retune:
                instructions.extend(["Warning both look around points and combined data are mapped to " +
                                     "up and down arrows, consider not returning and plotting combined " +
                                     "data at the same time"])
        instructions.append(("W-key", "write a pdf of all resonators", 'purple'))

        return instructions

    def plot_instructions(self):
        instructions = self.instructions()
        self.ax_key.clear()
        steps_per_item = 1.5
        y_step = 0.9 / (steps_per_item * float(len(instructions)))
        y_now = 0.95
        for key_press, description, color in instructions:
            self.ax_key.text(0.4, y_now, key_press.center(13), color=text_color_matplotlib[color],
                             ha='right', va='center', size=self.key_font_size, weight="bold",
                             family='monospace', bbox=dict(color=color, ls='-', lw=2.0, ec='black'))
            self.ax_key.text(0.45, y_now, description, color='black',
                             ha='left', va='center', size=self.key_font_size - 2)
            y_now -= steps_per_item * y_step
        self.ax_key.set_xlim(0, 1)
        self.ax_key.set_ylim(0, 1)
        if self.lasso_mode:
            self.ax_key.set_title('Lasso-Selection')
        else:
            self.ax_key.set_title('Main Menu')
        plt.draw()

    def print_instructions(self):
        instructions = self.instructions()
        for key_press, description, color in instructions:
            if color == 'black':
                text_color = 'white'
            else:
                text_color = 'black'
            color_text = colorize_text(text=key_press.capitalize().center(20), style_text='bold',
                                       color_text=text_color, color_background=color)
            if self.verbose:
                print(f'{color_text} : {description}')

    def set_flag_index(self, flag_index: int = None):
        if flag_index is None:
            # get the next flag index
            self.flag_type_index = (self.flag_type_index + 1) % len(self.flags_types)
        else:
            self.flag_type_index = flag_index
        if self.verbose:
            print(f'\n Flag mode is now: "{self.flags_types[self.flag_type_index]}"')
        self.print_instructions()
        self.plot_instructions()
        self.refresh_plot()

    def get_flag_type(self):
        flag_type = self.flags_types[self.flag_type_index]
        if flag_type == 'other':
            # allow for custom flag types to be added
            flag_type_new = input("Enter flag type: ").lower().strip()
            if flag_type_new and flag_type_new not in self.flags_types:
                # to new resonator type
                self.flags_types.append(flag_type_new)
                self.all_flags.add(flag_type_new)
                self.set_flag_index(flag_index=len(self.flags_types) - 1)
                flag_type = flag_type_new
        return flag_type

    def on_key_press(self, event):
        if platform.system().lower() == 'darwin' and event.key == 'a':
            self.shift_is_held = True
        elif platform.system().lower() != 'darwin' and event.key == 'shift':
            self.shift_is_held = True
        # items on all menus
        if event.key == 'right':
            if self.plot_index == self.res_indexes[-1]:
                self.plot_index = self.res_indexes[0]
            else:
                self.plot_index = self.res_indexes[np.where(self.res_indexes == self.plot_index)[-1] + 1][-1]
            self.refresh_plot()

        elif event.key == 'left':
            if self.plot_index == self.res_indexes[0]:
                self.plot_index = self.res_indexes[-1]
            else:
                self.plot_index = self.res_indexes[np.where(self.res_indexes == self.plot_index)[-1] - 1][0]
            self.refresh_plot()

        elif event.key == 'up':
            if self.look_around != self.chan_freqs.shape[0] // 2:
                self.look_around = self.look_around + 1
                self.update_min_index()
                if self.retune:
                    self.combined_data = np.expand_dims(self.min_index, 1)
            if self.combined_data is not None:
                if self.combined_data_index == self.combined_data.shape[1] - 1:
                    self.combined_data_index = 0
                else:
                    self.combined_data_index = self.combined_data_index + 1
            self.refresh_plot()

        elif event.key == 'down':
            if self.look_around != 1:
                self.look_around = self.look_around - 1
                self.update_min_index()
                if self.retune:
                    self.combined_data = np.expand_dims(self.min_index, 1)
            if self.combined_data is not None:
                if self.combined_data_index == 0:
                    self.combined_data_index = self.combined_data.shape[1] - 1
                else:
                    self.combined_data_index = self.combined_data_index - 1
            self.refresh_plot()
        # Writing an output file
        elif event.key == 'w':
            if self.verbose:
                print("saving to pdf")
            filename = input("enter filename for pdf: ")
            if filename == '':
                filename = "res_plots.pdf"
            elif filename[-4:] != '.pdf':
                filename = filename + '.pdf'
            self.make_pdf(filename)

        # Flagging and removing interactions
        elif not self.lasso_mode:
            if event.key == 'f':
                current_flags = self.flags[self.plot_index]
                if current_flags:
                    if self.verbose:
                        print(f"Res Index {self.plot_index} current flags: {current_flags}")
                # stage the selected point for flagging
                flag_type = self.get_flag_type()
                # stage the selected point for flagging
                if self.plot_index not in self.res_indexes_staged.keys():
                    self.res_indexes_staged[self.plot_index] = set()
                self.res_indexes_staged[self.plot_index].add(flag_type)
                # reset the plot
                self.plot_staged_for_removal(ax_combined=self.ax_combined)
                self.refresh_plot(autoscale=False)
                if self.verbose:
                    print(f'Res Index {self.plot_index} staged flags: {sorted(self.res_indexes_staged[self.plot_index])}')
                committed_flags = self.flags[self.plot_index]
            elif event.key == 'd':
                # cycle the flag mode by one type
                self.set_flag_index()
            elif event.key == 'e' or event.key == 't':
                # reset the plot instructions
                self.plot_instructions()
                if self.verbose:
                    print('\nMain Menu')
                self.print_instructions()
                if event.key == 't':
                    # save what is staged
                    for res_index in self.res_indexes_staged.keys():
                        for flag in sorted(self.res_indexes_staged[res_index]):
                            if flag == 'remove':
                                self.res_indexes_removed.add(res_index)
                            if flag == 'no-res':
                                self.res_indexes_removed.add(res_index)
                                self.flags[res_index].append(flag)
                            else:
                                self.flags[res_index].append(flag)
                    # reset the combined plot
                    self.combined_plot(ax_combined=self.ax_combined)
                self.res_indexes_staged = {}
                # this removes the staged points (red "X"s) from the plot
                self.plot_staged_for_removal(ax_combined=self.ax_combined)
                self.refresh_plot()
            elif event.key == 'z':
                flag_type = self.get_flag_type()
                if self.plot_index in self.res_indexes_staged:
                    flags_staged = self.res_indexes_staged[self.plot_index]
                    if flag_type in flags_staged:
                        # clear the staged point
                        flags_staged.remove(flag_type)
                        if not flags_staged:
                            del self.res_indexes_staged[self.plot_index]
                        self.plot_staged_for_removal(ax_combined=self.ax_combined)
                        self.refresh_plot()
                    else:
                        if self.verbose:
                            print(f"Res-Index {self.plot_index} not staged for flag type {flag_type}")
                else:
                    if self.verbose:
                        print(f"Res-Index {self.plot_index} not staged.")
            # Lasso selection
            elif event.key == 'b':
                self.lasso_mode = True
                self.plot_instructions()
                if self.verbose:
                    print('\nLasso Mode')
                self.print_instructions()
                self.ax_combined.set_title("Click and Drag to select points with a lasso (you click and draw on the plot).",
                                           size=16, color="firebrick", family="monospace")
                if self.pop_up_text is not None:
                    self.pop_up_text.remove()
                    self.pop_up_text = None
                self.popup_lasso()
                self.fig.canvas.draw()
                # lasso selector
                self.selector = SelectFromCollection(ax=self.ax_combined, collection=self.combined_data_points)
                self.fig.canvas.mpl_connect("key_press_event", self.accept_lasso)
                self.lasso_start_time = time.time()

    def accept_lasso(self, event_lasso):
        if self.lasso_mode:
            if event_lasso.key == "enter" or (event_lasso.key == "b" and time.time() > self.lasso_start_time + 1.0):
                if event_lasso.key == "enter":
                    selected_indexes = [int(index_as_float) for index_as_float
                                        in self.selector.xys[self.selector.ind][:, 0]]
                    flag_type = self.get_flag_type()
                    for index in selected_indexes:
                        if index not in self.res_indexes_staged:
                            self.res_indexes_staged[index] = set()
                        self.res_indexes_staged[index].add(flag_type)
                self.lasso_mode = False
                self.plot_staged_for_removal(ax_combined=self.ax_combined)
                self.selector.disconnect()
                self.selector = None
                self.refresh_plot()
                self.plot_instructions()
                if self.verbose:
                    print('\nMain Menu')
                self.print_instructions()

    def on_key_release(self, event):
        # windows or mac
        if platform.system() == 'Darwin':
            if event.key == 'a':
                self.shift_is_held = False
        else:
            if event.key == 'shift':
                self.shift_is_held = False

    def onClick(self, event):
        if self.combined_data is not None:
            if event.dblclick:
                # get the radius of each point from the click
                x_data_coords = self.res_indexes - event.xdata
                x_data_min, x_dat_max = self.ax_combined.get_xlim()
                x_data_range = x_dat_max - x_data_min
                x_norm_coords = x_data_coords / x_data_range
                x_yratio_coords = x_norm_coords * self.combined_x_width_over_y_height
                y_data_coords = self.combined_values_this_index - event.ydata
                y_data_min, y_dat_max = self.ax_combined.get_ylim()
                y_data_range = y_dat_max - y_data_min
                y_norm_coords = y_data_coords / y_data_range
                radius_array = np.sqrt(x_yratio_coords ** 2 + y_norm_coords ** 2)
                self.plot_index = self.res_indexes[np.argmin(radius_array)]
                self.refresh_plot(autoscale=False)
                return
        elif event.button == 3:
            if self.shift_is_held:
                if self.verbose:
                    print("overriding point selection", event.xdata)
                    # print(self.chan_freqs[:,self.plot_index][50])
                    # print((self.res_index_override == self.plot_index).any())
                if (self.res_index_override == self.plot_index).any():
                    replace_index = np.argwhere(
                        self.res_index_override == self.plot_index)[0][0]
                    new_freq = np.argmin(
                        np.abs(event.xdata - self.chan_freqs[:, self.plot_index] / 10 ** 6))
                    self.override_freq_index[replace_index] = np.int(new_freq)

                else:
                    self.res_index_override = np.append(
                        self.res_index_override, np.int(np.asarray(self.plot_index)))
                    # if self.verbose:
                    #     print(self.res_index_override)
                    new_freq = np.argmin(
                        np.abs(event.xdata - self.chan_freqs[:, self.plot_index] / 10 ** 6))
                    # if self.verbose:
                    #     print("new index is ",new_freq)
                    self.override_freq_index = np.append(
                        self.override_freq_index, np.int(np.asarray(new_freq)))
                    # if self.verbose:
                    #     print(self.override_freq_index)
                self.update_min_index()
                self.refresh_plot()

    def make_pdf(self, filename):
        pdf_pages = PdfPages(filename)
        for i in tqdm.tqdm(range(0, self.chan_freqs.shape[1]), ascii=True):
            self.plot_index = i
            self.refresh_plot()
            pdf_pages.savefig(self.fig)
        pdf_pages.close()


def tune_kids(f, z, find_min=True, interactive=True, **kwargs):
    # f and z should have shape (npts_sweep,n_res)
    # iq_dict = read_iq_sweep(filename)
    if "look_around" in kwargs:
        print("you are using " +
              str(kwargs['look_around']) + " look around points")
        look_around = kwargs['look_around']
    else:
        look_around = f.shape[0] // 2
    if find_min:  # fine the minimum
        print("centering on minimum")
        print(f.shape)
        if interactive:
            ip = InteractivePlot(f, z, look_around)
            print(ip.min_index.shape)
            print(ip.res_index_override.shape)
            for i in range(0, len(ip.res_index_override)):
                ip.min_index[ip.res_index_override[i]
                ] = ip.override_freq_index[i]
            new_freqs = f[(ip.min_index, np.arange(0, f.shape[1]))]
        else:
            min_index = np.argmin(np.abs(z) ** 2, axis=0)
            new_freqs = f[(min_index, np.arange(0, f.shape[1]))]
    else:  # find the max of dIdQ
        print("centering on max dIdQ")
        if interactive:
            ip = InteractivePlot(f, z, look_around, find_min=False)
            for i in range(0, len(ip.res_index_override)):
                ip.min_index[ip.res_index_override[i]
                ] = ip.override_freq_index[i]
            new_freqs = f[(ip.min_index, np.arange(0, f.shape[1]))]
        else:
            min_index = find_max_didq(z, look_around)
            new_freqs = f[(min_index, np.arange(0, f.shape[1]))]
    return new_freqs


class InteractivePowerTuningPlot(object):
    """
    special interactive plot for tune the readout power of resonators based on fist to their non-linearity parameter
    f (frequencies of iq_sweep shape n_pts_iq_sweep x n_res x (optionally n_powers)
    z complex number where i is real and q is imaginary part shape n_pts_iq_sweep x n_res,  n_powers
    fitted_a_mag from magnitude fits
    fitted_a_iq from iq fits (one of two fitted as must be provided
    attn_levels list or array of power levels
    desired_a the desired non-linearity parameter to have powers set to
    Optional
    z_fit_mag data to display as the fit
    z_fit_iq data to display as the fit
    """

    def __init__(self, f, z, attn_levels, fitted_a_mag=None, fitted_a_iq=None, desired_a=0.5, z_fit_mag=None,
                 z_fit_iq=None):
        self.attn_levels = np.asarray(attn_levels)
        self.bif_level = desired_a
        self.z = z
        self.Is = np.real(z)
        self.Qs = np.imag(z)
        self.z_fit_mag = z_fit_mag
        self.z_fit_iq = z_fit_iq
        if self.z_fit_iq is not None:
            self.Is_fit_iq = np.real(z_fit_iq)
            self.Qs_fit_iq = np.imag(z_fit_iq)
        self.fitted_a_mag = fitted_a_mag
        self.fitted_a_iq = fitted_a_iq
        if len(f.shape) > 2:
            self.f = f
        else:
            self.f = np.zeros(self.z.shape)
            for k in range(0, self.z.shape[1]):
                self.f[:, k, :] = self.f
        self.chan_freqs = f
        # self.targ_size = self.chan_freqs.shape[0]
        self.plot_index = 0
        self.power_index = 0
        self.res_index_override = np.asarray((), dtype=np.int16)
        self.override_freq_index = np.asarray((), dtype=np.int16)
        self.shift_is_held = False
        if fitted_a_mag is not None:
            self.bif_levels_mag = np.zeros(fitted_a_mag.shape[0])
        if fitted_a_iq is not None:
            self.bif_levels_iq = np.zeros(fitted_a_iq.shape[0])
        # here we compute our best guess for the proper power level from the fit data
        for i in range(0, self.f.shape[1]):
            if fitted_a_mag is not None:
                try:
                    first_bif_mag = np.where(fitted_a_mag[i, :] > bif_level)[0][0]
                except:
                    first_bif_mag = fitted_a_mag.shape[1] - 1
                if first_bif_mag == 0:
                    first_bif_mag = 1
                interp_mag = interpolate.interp1d(-self.attn_levels[0:first_bif_mag + 1],
                                                  fitted_a_mag[i, :][0:first_bif_mag + 1])
                powers_mag = np.linspace(-self.attn_levels[0] + .01, -self.attn_levels[first_bif_mag] - 0.01, 1000)
                bifurcation_mag = powers_mag[
                    np.argmin(np.abs(interp_mag(powers_mag) - np.ones(len(powers_mag)) * self.bif_level))]
                self.bif_levels_mag[i] = bifurcation_mag
            if fitted_a_iq is not None:
                try:
                    first_bif_iq = np.where(fitted_a_iq[i, :] > bif_level)[0][0]
                except:
                    first_bif_iq = self.fitted_a_iq.shape[1] - 1
                if first_bif_iq == 0:
                    first_bif_iq = 1
                interp_iq = interpolate.interp1d(-self.attn_levels[0:first_bif_iq + 1],
                                                 fitted_a_iq[i, :][0:first_bif_iq + 1])
                powers_iq = np.linspace(-self.attn_levels[0] + .01, -self.attn_levels[first_bif_iq] - 0.01, 1000)
                bifurcation_iq = powers_iq[
                    np.argmin(np.abs(interp_iq(powers_iq) - np.ones(len(powers_iq)) * self.bif_level))]
                self.bif_levels_iq[i] = bifurcation_iq

        if self.fitted_a_mag is not None:
            self.bif_levels = copy.deepcopy(self.bif_levels_mag)
        else:
            try:
                self.bif_levels = copy.deepcopy(self.bif_levels_iq)
            except:
                print("you must supply fits to the non-linearity parameter one of two fit variables")
                return
        self.power_index = np.argmin(np.abs(self.bif_levels[self.plot_index] + self.attn_levels))
        self.fig = plt.figure(1, figsize=(13, 10))
        self.ax = self.fig.add_subplot(221)
        self.ax.set_ylabel("Power (dB)")
        self.ax.set_xlabel("Frequecy (MHz)")
        self.ax2 = self.fig.add_subplot(222)
        self.ax2.set_ylabel("Q")
        self.ax2.set_xlabel("I")
        self.ax3 = self.fig.add_subplot(212)
        self.ax3.set_xlabel("Power level")
        self.ax3.set_ylabel("Nolinearity parameter a")
        self.fig.canvas.mpl_connect('key_press_event', self.on_key_press)
        self.fig.canvas.mpl_connect('key_release_event', self.on_key_release)
        self.fig.canvas.mpl_connect('button_press_event', self.onClick)
        if z_fit_mag is not None:
            self.l1_fit, = self.ax.plot(self.chan_freqs[:, self.plot_index, self.power_index],
                                        10 * np.log10(self.z_fit_mag[:, self.plot_index, self.power_index]),
                                        label="Fit")
        self.l1, = self.ax.plot(self.chan_freqs[:, self.plot_index, self.power_index],
                                20 * np.log10(np.abs(self.z[:, self.plot_index, self.power_index])), 'o', mec="k",
                                label="Data")
        self.ax.legend()
        if self.z_fit_iq is not None:
            self.l2_fit, = self.ax2.plot(self.Is_fit_iq[:, self.plot_index, self.power_index],
                                         self.Qs_fit_iq[:, self.plot_index, self.power_index], label="Fit")
        self.l2, = self.ax2.plot(self.Is[:, self.plot_index, self.power_index],
                                 self.Qs[:, self.plot_index, self.power_index],
                                 'o', mec="k", label="Data")
        self.ax2.legend()
        if self.fitted_a_mag is not None:
            self.l3, = self.ax3.plot(-self.attn_levels, self.fitted_a_mag[self.plot_index, :], color='r',
                                     label="Mag fit")
        if self.fitted_a_iq is not None:
            self.l4, = self.ax3.plot(-self.attn_levels, self.fitted_a_iq[self.plot_index, :], color='g', label="IQ fit")
        self.l8, = self.ax3.plot((self.bif_levels[self.plot_index], self.bif_levels[self.plot_index]), (0, 1),
                                 "--", color='m', linewidth=3, label="Power pick")
        self.l5, = self.ax3.plot((-self.attn_levels[self.power_index], -self.attn_levels[self.power_index]),
                                 (0, 1), "--", color='k', label="Current power")
        if self.fitted_a_mag is not None:
            self.l6, = self.ax3.plot((self.bif_levels_mag[self.plot_index], self.bif_levels_mag[self.plot_index]),
                                     (0, self.bif_level), "--", color='r')
        if self.fitted_a_iq is not None:
            self.l7, = self.ax3.plot((self.bif_levels_iq[self.plot_index], self.bif_levels_iq[self.plot_index]),
                                     (0, self.bif_level), "--", color='g')
        self.ax3.legend()
        self.ax3.set_ylim(0, 1)
        self.ax.set_title("Resonator index " + str(self.plot_index))
        self.ax2.set_title("Power level " + str(-self.attn_levels[self.power_index]))
        print("")
        print("Interactive Power Tuning Activated")
        print("Use left and right arrows to switch between resonators")
        print("Use the up and down arrows to change between power levels")
        print("Hold shift and right click on the bottom plot to override picked power level")
        print("or hold shift and press enter to override picked power level to the current plotted power level")
        plt.show(block=True)

    def refresh_plot(self):
        self.l1.set_data(self.chan_freqs[:, self.plot_index, self.power_index],
                         10 * np.log10(self.Is[:, self.plot_index, self.power_index] ** 2 + self.Qs[:, self.plot_index,
                                                                                            self.power_index] ** 2))
        if self.z_fit_mag is not None:
            self.l1_fit.set_data(self.chan_freqs[:, self.plot_index, self.power_index],
                                 10 * np.log10(self.z_fit_mag[:, self.plot_index, self.power_index]))
        self.ax.relim()
        self.ax.autoscale()
        self.ax.set_title("Resonator index " + str(self.plot_index))
        self.ax2.set_title("Power level " + str(-self.attn_levels[self.power_index]))
        self.l2.set_data((self.Is[:, self.plot_index, self.power_index], self.Qs[:, self.plot_index, self.power_index]))
        if self.z_fit_iq is not None:
            self.l2_fit.set_data((self.Is_fit_iq[:, self.plot_index, self.power_index],
                                  self.Qs_fit_iq[:, self.plot_index, self.power_index]))
        if self.fitted_a_mag is not None:
            self.l3.set_data(-self.attn_levels, self.fitted_a_mag[self.plot_index, :])
        self.l4.set_data(-self.attn_levels, self.fitted_a_iq[self.plot_index, :])
        self.l5.set_data((-self.attn_levels[self.power_index], -self.attn_levels[self.power_index]), (0, 1))
        if self.fitted_a_mag is not None:
            self.l6.set_data((self.bif_levels_mag[self.plot_index], self.bif_levels_mag[self.plot_index]),
                             (0, self.bif_level))
        if self.fitted_a_iq is not None:
            self.l7.set_data((self.bif_levels_iq[self.plot_index], self.bif_levels_iq[self.plot_index]),
                             (0, self.bif_level))

        self.l8.set_data((self.bif_levels[self.plot_index], self.bif_levels[self.plot_index]), (0, 1))
        self.ax2.relim()
        self.ax2.autoscale()
        plt.draw()

    def on_key_press(self, event):
        # print(event.key) #for debugging
        if event.key == 'right':
            if self.plot_index != self.chan_freqs.shape[1] - 1:
                self.plot_index = self.plot_index + 1
                # snap to the automated choice in power level
                self.power_index = np.argmin(np.abs(self.bif_levels[self.plot_index] + self.attn_levels))
                self.refresh_plot()

        if event.key == 'left':
            if self.plot_index != 0:
                self.plot_index = self.plot_index - 1
                # snap to the automated choice in power level
                self.power_index = np.argmin(np.abs(self.bif_levels[self.plot_index] + self.attn_levels))
                self.refresh_plot()

        if event.key == 'up':
            if self.power_index != self.Is.shape[2] - 1:
                self.power_index = self.power_index + 1
            self.refresh_plot()

        if event.key == 'down':
            if self.power_index != 0:
                self.power_index = self.power_index - 1
            self.refresh_plot()

        if event.key == 'shift':
            self.shift_is_held = True
            # print("shift pressed") #for debugging

        if event.key == 'enter':  # is this still needed on any os?
            # print("enter pressed")
            if self.shift_is_held:
                self.bif_levels[self.plot_index] = -self.attn_levels[self.power_index]
                self.refresh_plot()

        if event.key == 'shift+enter':
            self.bif_levels[self.plot_index] = -self.attn_levels[self.power_index]
            self.refresh_plot()

    def on_key_release(self, event):
        if event.key == "shift":
            self.shift_is_held = False
            # print("shift released") #for debugging

    def onClick(self, event):
        if event.button == 3:
            if self.shift_is_held:
                print("overiding point selection", event.xdata)
                self.bif_levels[self.plot_index] = event.xdata
                self.refresh_plot()


def tune_resonance_power(f, z, attn_levels, fitted_a_mag=None, fitted_a_iq=None, desired_a=0.5, z_fit_mag=None,
                         z_fit_iq=None):
    # Call the interactive plot
    ip = InteractivePowerTuningPlot(f, z, attn_levels, fitted_a_mag=fitted_a_mag,
                                    fitted_a_iq=fitted_a_iq, desired_a=desired_a,
                                    z_fit_mag=z_fit_mag, z_fit_iq=z_fit_iq)

    picked_power_levels = ip.bif_levels + attn_levels[-1]
    # plot the resultant picked power levels and the corresponding transfer function
    plt.figure(1, figsize=(12, 10))
    plt.subplot(211)
    plt.title("Bifurcation power")
    plt.plot(picked_power_levels, 'o', color='g', label="Picked Valuse")
    plt.xlabel("Resonator index")
    plt.ylabel("Power at a = " + str(desired_a))
    plt.legend(loc=4)

    normalizing_amplitudes = np.sqrt(
        10 ** ((ip.bif_levels + attn_levels[-1]) / 10) / np.min(10 ** ((ip.bif_levels + attn_levels[-1]) / 10)))
    plt.subplot(212)
    plt.title("Transfer function")
    plt.plot(normalizing_amplitudes, 'o')
    plt.xlabel("Resonator index")
    plt.ylabel("Voltage factor")

    # plt.savefig("power_sweep_results_"+fine_names[0].split('/')[-1]+".pdf")

    # save the output
    # np.savetxt("bif_levels_mag_"+fine_names[0].split('/')[-1]+".csv",ip.bif_levels_mag+attn_levels[-1])
    # np.savetxt("bif_levels_iq_"+fine_names[0].split('/')[-1]+".csv",ip.bif_levels_iq+attn_levels[-1])
    # np.save("trans_func_"+fine_names[0].split('/')[-1],np.sqrt(10**((ip.bif_levels+attn_levels[-1])/10)/np.min(10**((ip.bif_levels+attn_levels[-1])/10))))

    plt.show()
    return picked_power_levels, normalizing_amplitudes




from PyDAQmx import *
import matplotlib.pyplot as plt
import numpy as np
import time
from ctypes import *

class NIDAQ(object):
	
	# Constant Values // CHANGE ALL VALUES HERE
	set_sample_rate = 10000
	set_minimum_volt = -5
	set_maximum_volt = 5
	
	def __init__(self):
		# Initializing class values to user preference
		self.sample_rate = NIDAQ.set_sample_rate
		self.min_volt = NIDAQ.set_minimum_volt
		self.max_volt = NIDAQ.set_maximum_volt
		
		# Initializing Channel Names - CHANGE HERE
		chan1 = "Dev1/ai0"
		chan2 = "Dev1/ai1"
		chan3 = "Dev1/ai3"
		
		# Initializing array and AI input
		self.analog_input = Task()
		self.read = int32()
		
		# Channel names to pass to CreateAIVoltageChan
		self.nameof1CH = chan1
		self.nameof2CH = chan1 + ", " + chan2
		self.nameof3CH = chan1 + ", " + chan2 + ", " + chan3
		
	def stream_1ch(self, int_time):
		# This method opens 1 channel and streams that channel
		# Then returns the array of data
		
		# Gather necessary samples space for array
		timeOut = int_time + 100.0
		total_samples = int(self.sample_rate * int_time)
		self.data1 = np.zeros(((total_samples),), dtype=np.float64)
		
		# Open up the stream and gather data
		self.analog_input.CreateAIVoltageChan(self.nameof1CH,"",DAQmx_Val_Cfg_Default,self.min_volt,self.max_volt,DAQmx_Val_Volts,None)
		self.analog_input.CfgSampClkTiming("",int(self.sample_rate),DAQmx_Val_Rising,DAQmx_Val_FiniteSamps,total_samples)
		
		self.analog_input.StartTask()
		
		self.analog_input.ReadAnalogF64(total_samples,timeOut,DAQmx_Val_GroupByChannel,self.data1,total_samples,byref(self.read),None)
		self.analog_input.StopTask()
		self.analog_input.ClearTask()
		self.analog_input = Task()
		
		'''
		# Plotting onto screen
		plt.plot(self.data1)
		plt.show()
		'''
		return self.data1
		
	
	def stream_2ch(self, int_time):
		# This method opens 2 channels and streams the 2
		# Then returns 2 arrays
		
		# Gather necessary size for 2 array
		# Then initialize enough space for 2
		timeOut = int_time + 100.0
		total_samples = int(self.sample_rate * int_time)
		temp_data = np.zeros(((total_samples*2),), dtype=np.float64)
		self.data1 = np.zeros(((total_samples),), dtype=np.float64)
		self.data2 = np.zeros(((total_samples),), dtype=np.float64)

		
		# Open up the stream and gather data
		self.analog_input.CreateAIVoltageChan(self.nameof2CH,"",DAQmx_Val_Cfg_Default,self.min_volt,self.max_volt,DAQmx_Val_Volts,None)
		self.analog_input.CfgSampClkTiming("",int(self.sample_rate),DAQmx_Val_Rising,DAQmx_Val_FiniteSamps,total_samples)
		
		self.analog_input.StartTask()
		
		self.analog_input.ReadAnalogF64(total_samples,timeOut,DAQmx_Val_GroupByChannel,temp_data,total_samples*2,byref(self.read),None)
		
		self.analog_input.StopTask()
		self.analog_input.ClearTask()
		self.analog_input = Task()
		
		# Splitting the data into appropriate arrays
		self.data1 = temp_data[0:len(temp_data)/2]
		self.data2 = temp_data[len(temp_data)/2:]
		
		return self.data1, self.data2
		
	def stream_3ch(self, int_time):
		# This method opens 3 channels and streams the 3
		# Then returns 3 arrays
		
		timeOut = int_time + 100.0
		# Gather necessary size for 3 array
		# Then initialize enough space for 3
		total_samples = int(self.sample_rate * int_time)
		temp_data = np.zeros(((total_samples*3),), dtype=np.float64)
		self.data1 = np.zeros(((total_samples),), dtype=np.float64)
		self.data2 = np.zeros(((total_samples),), dtype=np.float64)
		self.data3 = np.zeros(((total_samples),), dtype=np.float64)
		
		# Open up the stream and gather data
		self.analog_input.CreateAIVoltageChan(self.nameof3CH,"",DAQmx_Val_Cfg_Default,self.min_volt,self.max_volt,DAQmx_Val_Volts,None)
		self.analog_input.CfgSampClkTiming("",int(self.sample_rate),DAQmx_Val_Rising,DAQmx_Val_FiniteSamps,total_samples)
		
		self.analog_input.StartTask()
		
		self.analog_input.ReadAnalogF64(total_samples,timeOut,DAQmx_Val_GroupByChannel,temp_data,total_samples*3,byref(self.read),None)
		
		# Splitting the data stored in temp_data into
		# appropriate arrays
		self.data1 = temp_data[0:len(temp_data)/3]
		self.data2 = temp_data[len(temp_data)/3:(len(temp_data)-(len(temp_data)/3))]
		self.data3 = temp_data[(len(temp_data)-(len(temp_data)/3)):]
		
		self.analog_input.StopTask()
		self.analog_input.ClearTask()
		self.analog_input = Task()
		
		return self.data1, self.data2, self.data3
	
	def average_1ch(self, int_time):
		# Returns average for data1 set

		total = np.mean(self.stream_1ch(int_time))
		
		return total
	
	def average_2ch(self, int_time):
		# Returns 2 averages, one for data1 set
		# and the other for data2 set
		total1, total2 = self.stream_2ch(int_time)
		average1 = np.mean(total1)
		average2 = np.mean(total2)
		
		return average1, average2
		
	def help(self):
		# This method is used to show user all of the
		# available functions for NIDAQ
		
		print("\t Available Functions")
		print("\t -------------------")
		print("\t 1) stream_1ch")
		print("\t 2) stream_2ch")
		print("\t 3) stream_3ch")
		print("\t 4) average_1ch")
		print("\t 5) average_2ch")
		